# - Imports
from __future__ import print_function

# Communication imports
from pyniryo2.robot_commander import RobotCommander
from pyniryo2.enums import RobotErrors
from pyniryo2.objects import PoseObject
from pyniryo2.utils import pose_dict_to_list

from .enums import ObjectShape, ObjectColor, ManageWorkspace
from .services import VisionServices
from .topics import VisionTopics

from pyniryo2.arm.arm import Arm
from pyniryo2.tool.tool import Tool


def object_pose_dict_to_pose_object(object_pose_dict):
    return PoseObject(*[object_pose_dict[axis] for axis in ["x", "y", "z", "roll", "pitch", "yaw"]])


class Vision(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client, arm=None, tool=None):
        """
        Vision robot functions

        Example: ::

            ros_instance = NiryoRos("10.10.10.10") # Hotspot
            vision_interface = Vision(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(Vision, self).__init__(client)

        self._services = VisionServices(self._client)
        self._topics = VisionTopics(self._client)

        if arm is None:
            self.__arm = Arm(client)
        else:
            self._check_instance(arm, Arm)
            self.__arm = arm

        if tool is None:
            self.__tool = Tool(client)
        else:
            self._check_instance(tool, Tool)
            self.__tool = tool

    @property
    def get_camera_intrinsics(self):
        """
        Get calibration object: camera intrinsics, distortions coefficients
        The topic return a namedtuple(intrinsics: list[list[float]], distortion: list[list[float]])

        Examples: ::

            vision.get_camera_intrinsics()
            vision.get_camera_intrinsics().value

            def camera_info_callback(camera_info):
                print(camera_info.intrinsics)
                print(camera_info.distortion)
                vision.get_camera_intrinsics.unsubscribe()

            vision.get_camera_intrinsics.subscribe(camera_info_callback)

        :rtype: NiryoTopic
        """
        return self._topics.camera_info_topic

    @property
    def get_img_compressed(self):
        """
        Get image from video stream in a compressed format.
        Use ``uncompress_image`` from the vision package to uncompress it

        Examples: ::

            import pyniryo

            img_compressed = vision.get_img_compressed()
            camera_info = vision.get_camera_intrinsics()
            img = pyniryo.uncompress_image(img_compressed)
            img = pyniryo.undistort_image(img, camera_info.intrinsics, camera_info.distortion)


        :return: string containing a JPEG compressed image
        :rtype: NiryoTopic
        """
        return self._topics.compressed_video_stream_topic

    @property
    def get_image_parameters(self):
        """
        Return the NiryoTopic to get last stream image parameters:
        Brightness factor, Contrast factor, Saturation factor.
        The topic return a namedtuple(brightness_factor: float, contrast_factor: float, saturation_factor: float)

        Brightness factor: How much to adjust the brightness. 0.5 will give a darkened image,
        1 will give the original image while 2 will enhance the brightness by a factor of 2.

        Contrast factor: A factor of 1 gives original image.
        Making the factor towards 0 makes the image greyer, while factor>1 increases the contrast of the image.

        Saturation factor: 0 will give a black and white image, 1 will give the original image while
        2 will enhance the saturation by a factor of 2.

        Examples: ::

            vision.get_image_parameters()
            vision.get_image_parameters.value

            def image_parameters_callback(image_parameters):
                print(image_parameters.brightness_factor)
                print(image_parameters.contrast_factor)
                print(image_parameters.saturation_factor)

                vision.get_image_parameters.unsubscribe()

            vision.get_image_parameters.subscribe(image_parameters_callback)


        :return: ImageParameters namedtuple containing the brightness factor, contrast factor and saturation factor.
        :rtype: NiryoTopic
        """
        return self._topics.video_stream_parameters_topic

    def set_brightness(self, brightness_factor):
        """
        Modify image brightness

        :param brightness_factor: How much to adjust the brightness. 0.5 will
            give a darkened image, 1 will give the original image while
            2 will enhance the brightness by a factor of 2.
        :type brightness_factor: float
        :rtype: None
        """
        self._check_not_instance(brightness_factor, bool)
        self._check_instance(brightness_factor, (int, float))
        req = self._services.get_image_parameter_request(brightness_factor)
        result = self._services.set_brightness_service.call(req)
        self._check_result_status(result)

    def set_contrast(self, contrast_factor):
        """
        Modify image contrast

        :param contrast_factor: A factor of 1 gives original image.
            Making the factor towards 0 makes the image greyer, while factor>1 increases the contrast of the image.
        :type contrast_factor: float
        :rtype: None
        """
        self._check_not_instance(contrast_factor, bool)
        self._check_instance(contrast_factor, (int, float))
        req = self._services.get_image_parameter_request(contrast_factor)
        result = self._services.set_contrast_service.call(req)
        self._check_result_status(result)

    def set_saturation(self, saturation_factor):
        """
        Modify image saturation

        :param saturation_factor: How much to adjust the saturation. 0 will
            give a black and white image, 1 will give the original image while
            2 will enhance the saturation by a factor of 2.
        :type saturation_factor: float
        :rtype: None
        """
        self._check_not_instance(saturation_factor, bool)
        self._check_instance(saturation_factor, (int, float))
        req = self._services.get_image_parameter_request(saturation_factor)
        result = self._services.set_saturation_service.call(req)
        self._check_result_status(result)

    def get_target_pose_from_rel(self, workspace_name, height_offset, x_rel, y_rel, yaw_rel):
        """
        Given a pose (x_rel, y_rel, yaw_rel) relative to a workspace, this function
        returns the robot pose in which the current tool will be able to pick an object at this pose.

        The height_offset argument (in m) defines how high the tool will hover over the workspace. If height_offset = 0,
        the tool will nearly touch the workspace.

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param x_rel: x relative pose (between 0 and 1)
        :type x_rel: float
        :param y_rel: y relative pose (between 0 and 1)
        :type y_rel: float
        :param yaw_rel: Angle in radians
        :type yaw_rel: float

        :return: target_pose
        :rtype: PoseObject
        """
        self._check_type(workspace_name, str)
        self._check_range_belonging(x_rel, 0.0, 1.0)
        self._check_range_belonging(y_rel, 0.0, 1.0)

        height_offset, x_rel, y_rel, yaw_rel = self._map_list([height_offset, x_rel, y_rel, yaw_rel], float)

        req = self._services.get_target_pose_service_request(workspace_name, height_offset, x_rel, y_rel, yaw_rel)
        resp = self._services.get_target_pose_service.call(req)

        pose_array = pose_dict_to_list(resp["target_pose"])
        pose_object = PoseObject(*pose_array)
        return pose_object

    def get_target_pose_from_cam(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        First detects the specified object using the camera and then returns the robot pose in which the object can
        be picked with the current tool

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, ObjectShape, ObjectColor)
        """
        self._check_type(workspace_name, str)
        height_offset = self._transform_to_type(height_offset, float)
        self._check_enum_belonging(shape, ObjectShape)
        self._check_enum_belonging(color, ObjectColor)

        object_found, rel_pose, obj_shape, obj_color = self.detect_object(workspace_name, shape, color)

        if not object_found:
            obj_pose = PoseObject(0, 0, 0, 0, 0, 0)
        else:
            obj_pose = self.get_target_pose_from_rel(workspace_name, height_offset, rel_pose.x, rel_pose.y,
                                                     rel_pose.yaw)

        return object_found, obj_pose, obj_shape, obj_color

    def vision_pick(self, workspace_name, height_offset=0.0, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Picks the specified object from the workspace. This function has multiple phases: \n
        | 1. detect object using the camera
        | 2. prepare the current tool for picking
        | 3. approach the object
        | 4. move down to the correct picking pose
        | 5. actuate the current tool
        | 6. lift the object

        Example::

            robot = NiryoRobot(ip_address="x.x.x.x")
            robot.arm.calibrate_auto()
            robot.arm.move_pose(<observation_pose>)
            obj_found, shape_ret, color_ret = robot.vision.vision_pick(<workspace_name>,
                                                                height_offset=0.0,
                                                                shape=ObjectShape.ANY,
                                                                color=ObjectColor.ANY)

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        self._check_type(workspace_name, str)
        height_offset = self._transform_to_type(height_offset, float)
        self._check_enum_belonging(shape, ObjectShape)
        self._check_enum_belonging(color, ObjectColor)

        object_found, rel_pose, obj_shape, obj_color = self.detect_object(workspace_name, shape, color)
        if not object_found:
            return False, "", ""

        pick_pose = self.get_target_pose_from_rel(
            workspace_name, height_offset, rel_pose.x, rel_pose.y, rel_pose.yaw)
        approach_pose = self.get_target_pose_from_rel(
            workspace_name, height_offset + 0.05, rel_pose.x, rel_pose.y, rel_pose.yaw)

        self.__tool.release_with_tool()

        self.__arm.move_pose(approach_pose)
        self.__arm.move_pose(pick_pose)

        self.__tool.grasp_with_tool()

        self.__arm.move_pose(approach_pose)
        return True, obj_shape, obj_color

    def move_to_object(self, workspace_name, height_offset, shape, color):
        """
        Same as `get_target_pose_from_cam` but directly moves to this position

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param height_offset: offset between the workspace and the target height
        :type height_offset: float
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_shape, object_color
        :rtype: (bool, ObjectShape, ObjectColor)
        """
        obj_found, obj_pose, obj_shape, obj_color = self.get_target_pose_from_cam(workspace_name, height_offset, shape,
                                                                                  color)
        if not obj_found:
            return False, "", ""

        self.__arm.move_pose(obj_pose)
        return True, obj_shape, obj_color

    def detect_object(self, workspace_name, shape=ObjectShape.ANY, color=ObjectColor.ANY):
        """
        Detect object in workspace and return its pose and characteristics

        :param workspace_name: name of the workspace
        :type workspace_name: str
        :param shape: shape of the target
        :type shape: ObjectShape
        :param color: color of the target
        :type color: ObjectColor
        :return: object_found, object_pose, object_shape, object_color
        :rtype: (bool, PoseObject, str, str)
        """
        self._check_type(workspace_name, str)
        self._check_enum_belonging(shape, ObjectShape)
        self._check_enum_belonging(color, ObjectColor)

        ratio = self.get_workspace_ratio(workspace_name)
        req = self._services.obj_detection_rel_service_request(shape, color, ratio)
        resp = self._services.obj_detection_rel_service.call(req)

        obj_found = resp["status"] >= RobotErrors.SUCCESS.value
        if not obj_found:
            rel_pose = PoseObject(*(6 * [0.0]))
            shape = "ANY"
            color = "ANY"
        else:
            rel_pose = object_pose_dict_to_pose_object(resp["obj_pose"])
            shape = resp["obj_type"]
            color = resp["obj_color"]

        return obj_found, rel_pose, ObjectShape[shape], ObjectColor[color]

    # - Workspace
    def save_workspace_from_robot_poses(self, workspace_name, pose_origin, pose_2, pose_3, pose_4):
        """
        Save workspace by giving the poses of the robot to point its 4 corners
        with the calibration Tip. Corners should be in the good order.
        Markers' pose will be deduced from these poses

        Poses should be either a list [x, y, z, roll, pitch, yaw] or a PoseObject

        :param workspace_name: workspace name
        :type workspace_name: str
        :param pose_origin:
        :type pose_origin: Union[list[float], PoseObject]
        :param pose_2:
        :type pose_2: Union[list[float], PoseObject]
        :param pose_3:
        :type pose_3: Union[list[float], PoseObject]
        :param pose_4:
        :type pose_4: Union[list[float], PoseObject]

        :rtype: None
        """
        self._check_type(workspace_name, str)
        pose_list = [self._args_pose_to_list(pose) for pose in (pose_origin, pose_2, pose_3, pose_4)]
        req = self._services.add_workspace_from_poses_request(workspace_name, pose_list)
        _resp = self._services.manage_workspace_service.call(req)

    def save_workspace_from_points(self, workspace_name, point_origin, point_2, point_3, point_4):
        """
        Save workspace by giving the points of worskpace's 4 corners. Points are written as [x, y, z]
        Corners should be in the good order.

        :param workspace_name: workspace name
        :type workspace_name: str
        :param point_origin:
        :type point_origin: list[float]
        :param point_2:
        :type point_2: list[float]
        :param point_3:
        :type point_3: list[float]
        :param point_4:
        :type point_4: list[float]
        :rtype: None
        """
        self._check_type(workspace_name, str)
        point_list = [self._map_list(point, float) for point in (point_origin, point_2, point_3, point_4)]
        req = self._services.add_workspace_from_points_request(workspace_name, point_list)
        _resp = self._services.manage_workspace_service.call(req)

    def delete_workspace(self, workspace_name):
        """
        Delete workspace from robot's memory

        :param workspace_name: name of the saved workspace
        :type workspace_name: str
        :rtype: None
        """
        self._check_type(workspace_name, str)
        workspace = {"name": workspace_name}
        req = self._services.manage_workspace_service_request(ManageWorkspace.DELETE, workspace)
        _resp = self._services.manage_workspace_service.call(req)

    def get_workspace_poses(self, workspace_name):
        """
        Get the position of the tags of the workspace from the robot's memory.

        :param workspace_name:  name of the saved workspace
        :type workspace_name: str
        :return: List of the 4 workspace poses in the clockwise
        :rtype: list[PoseObject]
        """
        self._check_type(workspace_name, str)
        req = self._services.get_workspace_poses_service_request(workspace_name)
        resp = self._services.get_workspace_poses_service.call(req)
        return [PoseObject(*pose_dict_to_list(robot_pose)) for robot_pose in resp["poses"]]

    def get_workspace_ratio(self, workspace_name):
        """
        Get workspace ratio from robot's memory

        :param workspace_name: workspace name
        :type workspace_name: str
        :rtype: float
        """
        self._check_type(workspace_name, str)
        req = self._services.get_workspace_ratio_service_request(workspace_name)
        return self._services.get_workspace_ratio_service.call(req)["ratio"]

    def get_workspace_list(self):
        """
        Get list of workspaces' name store in robot's memory

        :rtype: list[str]
        """
        req = self._services.get_workspace_list_service_request()
        return self._map_list(self._services.get_workspace_list_service.call(req)["name_list"], str)



# - Imports
from __future__ import print_function

# Communication imports
from pyniryo2.robot_commander import RobotCommander

from pyniryo2.arm.arm import Arm
from pyniryo2.tool.tool import Tool
from pyniryo2.trajectories.trajectories import Trajectories


class PickPlace(RobotCommander):
    def __init__(self, client, arm=None, tool=None, trajectories=None):
        super(PickPlace, self).__init__(client)

        if arm is None:
            self.__arm = Arm(client)
        else:
            self._check_instance(arm, Arm)
            self.__arm = arm

        if tool is None:
            self.__tool = Tool(client)
        else:
            self._check_instance(tool, Tool)
            self.__tool = tool

        if trajectories is None:
            self.__trajectories = Trajectories(client)
        else:
            self._check_instance(trajectories, Trajectories)
            self.__trajectories = trajectories

    def pick_from_pose(self, *args):
        """
        Execute a picking from a pose.

        A picking is described as : \n
        | * going over the object
        | * going down until height = z
        | * grasping with tool
        | * going back over the object

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        pick_pose = self._args_pose_to_pose_object(*args)
        pick_pose_high = pick_pose.copy_with_offsets(z_offset=0.05)

        self.__tool.release_with_tool()
        self.__trajectories.execute_trajectory_from_poses([pick_pose_high, pick_pose], dist_smoothing=0.0)
        self.__tool.grasp_with_tool()

        return self.__arm.move_pose(pick_pose_high)

    def place_from_pose(self, *args):
        """
        Execute a placing from a position.

        A placing is described as : \n
        | * going over the place
        | * going down until height = z
        | * releasing the object with tool
        | * going back over the place

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """

        place_pose = self._args_pose_to_pose_object(*args)
        place_pose_high = place_pose.copy_with_offsets(z_offset=0.05)

        self.__trajectories.execute_trajectory_from_poses([place_pose_high, place_pose], dist_smoothing=0.0)
        self.__tool.release_with_tool()

        return self.__arm.move_pose(place_pose_high)

    def pick_and_place(self, pick_pose, place_pose, dist_smoothing=0.0):
        """
        Execute a pick then a place

        :param pick_pose: Pick Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type pick_pose: Union[list[float], PoseObject]
        :param place_pose: Place Pose : [x, y, z, roll, pitch, yaw] or PoseObject
        :type place_pose: Union[list[float], PoseObject]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        pick_pose_object = self._args_pose_to_pose_object(pick_pose)
        place_pose_object = self._args_pose_to_pose_object(place_pose)

        pick_pose_high = pick_pose_object.copy_with_offsets(z_offset=0.05)
        place_pose_high = place_pose_object.copy_with_offsets(z_offset=0.05)

        self.__tool.release_with_tool()

        self.__trajectories.execute_trajectory_from_poses([pick_pose_high, pick_pose_object], dist_smoothing)
        self.__tool.grasp_with_tool()

        self.__trajectories.execute_trajectory_from_poses([pick_pose_high, place_pose_high, place_pose_object],
                                                          dist_smoothing)

        self.__tool.release_with_tool()

        return self.__arm.move_pose(place_pose_high)



#!/usr/bin/env python
import unittest
from threading import Event
import numpy as np

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.niryo_ros import NiryoRos
from pyniryo2.objects import PoseObject

from pyniryo2.frames.frames import Frames
from pyniryo2.arm.arm import Arm

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_creation_edition_frame",
              "test_move_in_frame",
              "test_deletion"
              ]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.frames = Frames(cls.client)
        cls.arm = Arm(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.arm.go_to_sleep()
        #cls.client.close()

    @staticmethod
    def assertAlmostEqualVector(a, b, decimal=1):
        np.testing.assert_almost_equal(a, b, decimal)

# noinspection PyTypeChecker


class TestFrames(BaseTest):
    robot_poses = [[
        [0.2, 0.2, 0.1, 0, 0, 0],
        [0.4, 0.3, 0.1, 0, 0, 0],
        [0.3, 0.4, 0.1, 0, 0, 0]
    ],
        [
        [-0.2, -0.2, 0.1, 0, 0, 0],
        [-0.4, -0.3, 0.1, 0, 0, 0],
        [-0.3, -0.4, 0.1, 0, 0, 0],
    ]
    ]

    robot_point = [[
        [-0.2, 0.2, 0.1],
        [0.4, 0.3, 0],
        [0.3, 0.4, 0]
    ],
        [
        [0.2, -0.2, 0.1],
        [-0.4, -0.3, 0],
        [-0.3, -0.4, 0]
    ]
    ]

    def test_creation_edition_frame(self):
        base_dict =  self.frames.get_saved_dynamic_frame_list()
        new_base_dict = base_dict.copy()

        # Create frame by poses
        list_saved = []
        for i in range(4):
            if (i < 2):
                # Test creation by poses
                new_name = 'unitTestFramePose_{:03d}'.format(i)
                new_desc = 'descTestFramePose_{:03d}'.format(i)
                pose_o = self.robot_poses[i][0]
                pose_x = self.robot_poses[i][1]
                pose_y = self.robot_poses[i][2]
                self.assertIsNone(self.frames.save_dynamic_frame_from_poses(new_name, new_desc, pose_o, pose_x, pose_y))

                # Test edition
                new_edit_name = 'unitEditTestFramePose_{:03d}'.format(i)
                new_edit_desc = 'descEditTestFramePose_{:03d}'.format(i)
                self.assertIsNone(self.frames.edit_dynamic_frame(new_name, new_edit_name, new_edit_desc))
                self.assertEqual(self.frames.get_saved_dynamic_frame(new_edit_name).description, new_edit_desc)

                with self.assertRaises(RobotCommandException):
                    self.frames.get_saved_dynamic_frame(0)


                new_base_dict[new_edit_name] = new_edit_desc

                self.assertEqual(self.frames.get_saved_dynamic_frame_list(), new_base_dict)

                with self.assertRaises(RobotCommandException):
                    self.frames.save_dynamic_frame_from_poses(0, "unittest", pose_o, pose_x, pose_y)

                with self.assertRaises(RobotCommandException):
                    self.frames.save_dynamic_frame_from_points(0, "unittest", pose_o, pose_x, pose_y)

                with self.assertRaises(RobotCommandException):
                    self.frames.edit_dynamic_frame("unitTestFramePose_000", 0, 0)

            else:
                # Test creation by points
                new_name = 'unitTestFramePose_{:03d}'.format(i)
                new_desc = 'descTestFramePose_{:03d}'.format(i)
                point_o = self.robot_point[2-i][0]
                point_x = self.robot_point[2-i][1]
                point_y = self.robot_point[2-i][2]
                self.assertIsNone(self.frames.save_dynamic_frame_from_points(new_name, new_desc, point_o, point_x, point_y))

                # Test edition
                new_edit_name = 'unitEditTestFramePose_{:03d}'.format(i)
                new_edit_desc = 'descEditTestFramePose_{:03d}'.format(i)
                self.assertIsNone(self.frames.edit_dynamic_frame(new_name, new_edit_name, new_edit_desc))
                self.assertEqual(self.frames.get_saved_dynamic_frame(new_edit_name)[1], new_edit_desc)

                new_base_dict[new_edit_name] = new_edit_desc

                self.assertEqual(self.frames.get_saved_dynamic_frame_list(), new_base_dict)

    def test_move_in_frame(self):
        # Move frame 000
        pose0 = (0, 0, 0, 0, 1.57, 0)
        self.assertIsNone(self.arm.move_pose(pose0, "unitEditTestFramePose_000"))
        self.assertIsNone(self.arm.move_linear_pose((0.05, 0.05, 0.05, 0, 1.57, 0), "unitEditTestFramePose_000"))

        # Move frame 001
        pose1 = PoseObject(0, 0, 0, 0, 1.57, 0)
        self.assertIsNone(self.arm.move_pose(pose1, "unitEditTestFramePose_001"))
        self.assertIsNone(self.arm.move_linear_pose((0.05, 0.05, 0.05, 0, 1.57, 0), "unitEditTestFramePose_001"))

        # Move frame 002
        pose2 = (0, 0, 0, 0, 1.57, 0)
        self.assertIsNone(self.arm.move_pose(pose2, "unitEditTestFramePose_002"))
        self.assertIsNone(self.arm.move_relative([0.05, 0.05, 0.05, 0.1, 0.1, 0.1], "unitEditTestFramePose_002"))
        self.assertIsNone(self.arm.move_linear_relative([-0.05, -0.05, -0.05, 0, 0, 0], "unitEditTestFramePose_002"))

        # Move frame 003
        pose3 = PoseObject(0, 0, 0, 0, 1.57, 0)
        self.assertIsNone(self.arm.move_pose(pose3, "unitEditTestFramePose_003"))
        self.assertIsNone(self.arm.move_relative([0.05, 0.05, 0.05, 0.1, 0.1, 0.1], "unitEditTestFramePose_003"))
        self.assertIsNone(self.arm.move_linear_relative([-0.05, -0.05, -0.05, 0, 0, 0], "unitEditTestFramePose_003"))

        # Test default world
        self.assertIsNone(self.arm.move_linear_relative([0.1, 0.1, 0.1, 0, 0, 0]))
        self.assertIsNone(self.arm.move_linear_relative([0, 0, -0.1, 0, 0, 0]))

        with self.assertRaises(RobotCommandException):
            self.arm.move_relative([0.05, 0.05, 0.05, 0.1, 0.1, 0.1], 0)

        with self.assertRaises(RobotCommandException):
            self.arm.move_linear_relative([0.05, 0.05, 0.05, 0.1, 0.1, 0.1], 0)

    def test_deletion(self):
        base_dict = self.frames.get_saved_dynamic_frame_list()
        new_base_dict = base_dict.copy()

        for i in range(4):
            name_delete = 'unitEditTestFramePose_{:03d}'.format(i)
            self.assertIsNone(self.frames.delete_saved_dynamic_frame(name_delete))

            new_base_dict.pop(name_delete)

            self.assertEqual(self.frames.get_saved_dynamic_frame_list(), new_base_dict)


def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestFrames(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



from pyniryo2.enums import RobotErrors
from pyniryo2.exceptions import RobotCommandException
from pyniryo2.robot_commander import RobotCommander
from .enums import PinID, PinMode, PinState
from .services import IOServices
from .topics import IOTopics
from .objects import DigitalPinObject, AnalogPinObject


def check_ned2_version(func):
    """
    Decorator that check the robot version
    """

    def wrap(*args, **kwargs):
        robot_instance = args[0]
        if robot_instance.client.hardware_version != 'ned2':
            raise Exception("Error Code : BAD_HARDWARE_VERSION\n"
                            "Message : Wrong robot hardware version, feature only available on Ned2")

        return func(*args, **kwargs)

    return wrap


def check_ned_one_version(func):
    """
    Decorator that check the robot version
    """

    def wrap(*args, **kwargs):
        robot_instance = args[0]
        if robot_instance.client.hardware_version != 'ned2':
            raise Exception("Error Code : BAD_HARDWARE_VERSION\n"
                            "Message : Wrong robot hardware version, feature only available on Ned1 and One")

        return func(*args, **kwargs)

    return wrap


class IO(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        """
        IO robot functions

        Example: ::

            ros_instance = NiryoRos("10.10.10.10") # Hotspot
            io_interface = IO(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(IO, self).__init__(client)

        self._services = IOServices(self._client)
        self._topics = IOTopics(self._client)

    @property
    def digital_io_states(self):
        """
        Return the value of digital ios.

        :return: State, Name, Mode of the pin
        :rtype: DigitalPinObject
        """
        return self.get_digital_io_states()

    @property
    def get_digital_io_states(self):
        """
        Returns the io state client which can be used synchronously or asynchronously to obtain the io states.
        The io state client returns a list of DigitalPinObject.

        Examples: ::

            # Get last value
            io.get_digital_io_states()
            io.get_digital_io_states.value

            # Subscribe a callback
            def io_callback(io_state):
                print io_state

            io.get_digital_io_states.subscribe(io_callback)
            io.get_digital_io_states.unsubscribe()

        :return: io state topic instance
        :rtype: NiryoTopic
        """
        return self._topics.digital_io_topic

    def get_digital_io_state(self, pin_id):
        """
        Return the value of a digital io.

        :type pin_id: PinID or str
        :return: digital io object
        :rtype: DigitalPinObject
        """
        self._check_instance(pin_id, (PinID, str))
        states = self._topics.digital_io_topic()

        for state in states:
            if state.pin_id == pin_id or state.name == pin_id:
                return state

        raise RobotCommandException('IO {} not found'.format(pin_id))

    @check_ned_one_version
    def set_pin_mode(self, pin_id, pin_mode):
        """
        Set pin number pin_id to mode pin_mode

        Examples: ::

            io.set_pin_mode(PinID.GPIO_1A, PinMode.INPUT)
            io.set_pin_mode("1A", PinMode.OUTPUT)

        :param pin_id:
        :type pin_id: PinID or str
        :param pin_mode:
        :type pin_mode: PinMode
        :rtype: None
        """
        self._check_instance(pin_id, (PinID, str))
        self._check_enum_belonging(pin_mode, PinMode)

        req = self._services.set_io_mode_request(pin_id, pin_mode)
        resp = self._services.set_digital_io_mode_service.call(req)
        self._check_result_status(resp)

    def digital_write(self, pin_id, digital_state):
        """
        Set pin_id state to digital_state

        Examples: ::

            io.digital_write(PinID.GPIO_1A, PinState.HIGH)
            io.digital_write('1A', PinState.LOW)

        :param pin_id:
        :type pin_id: PinID or String
        :param digital_state:
        :type digital_state: PinState or bool
        :rtype: None
        """
        self._check_instance(pin_id, (PinID, str))
        self._check_instance(digital_state, (PinState, bool))

        req = self._services.set_io_state_request(pin_id, digital_state)
        resp = self._services.set_digital_io_state_service.call(req)

        self._check_result_status(resp)

    def digital_read(self, pin_id):
        """
        Return the value of a digital pin.

        Examples: ::

           io.set_pin_mode(PinID.GPIO_1A, PinMode.OUTPUT)
           io.digital_read(PinID.GPIO_1A)

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: bool
        """
        self._check_instance(pin_id, (PinID, str))

        req = self._services.get_io_request(pin_id)
        resp = self._services.get_digital_io_service.call(req)

        if resp["status"] < RobotErrors.SUCCESS.value:
            return None

        return resp["value"]

    @property
    def analog_io_states(self):
        """
        Return the value of analog ios.
        Only available on Ned2.

        :return: State, Name, Mode of the pin
        :rtype: AnalogPinObject
        """
        return self.get_digital_io_states()

    @property
    def get_analog_io_states(self):
        """
        Returns the io state client which can be used synchronously or asynchronously to obtain the io states.
        The io state client returns a list of AnalogPinObject.
        Only available on Ned2.

        Examples: ::

            # Get last value
            io.get_analog_io_states()
            io.get_analog_io_states.value

            # Subscribe a callback
            def io_callback(io_state):
                print io_state

            io.get_analog_io_states.subscribe(io_callback)
            io.get_analog_io_states.unsubscribe()

        :return: io state topic instance
        :rtype: NiryoTopic
        """
        return self._topics.analog_io_topic

    def get_analog_io_state(self, pin_id):
        """
        Return the value of an analog io.
        Only available on Ned2.

        :type pin_id: PinID or str
        :return: analog io object
        :rtype: AnalogPinObject
        """
        self._check_instance(pin_id, (PinID, str))
        states = self._topics.analog_io_topic()

        for state in states:
            if state.pin_id == pin_id or state.name == pin_id:
                return state

        raise RobotCommandException('Analog IO {} not found'.format(pin_id))

    def analog_write(self, pin_id, value):
        """
        Set pin_id state to digital_state.
        Only available on Ned2.

        Examples: ::

            io.digital_write(PinID.AO1, 0.0)
            io.digital_write('AO1', 5.0)

        :param pin_id:
        :type pin_id: PinID or String
        :param value: voltage
        :type value: float
        :rtype: None
        """
        self._check_instance(pin_id, (PinID, str))
        self._check_instance(value, (float, int))
        self._check_range_belonging(value, 0, 5)

        req = self._services.set_io_state_request(pin_id, value)
        resp = self._services.set_analog_io_state_service.call(req)

        self._check_result_status(resp)

    def analog_read(self, pin_id):
        """
        Read the value of an analog pin.
        Only available on Ned2.

        Examples: ::

           io.analog_read(PinID.AI1)
           io.analog_read('AI1')

        :param pin_id:
        :type pin_id: PinID or str
        :rtype: bool
        """
        self._check_instance(pin_id, (PinID, str))

        req = self._services.get_io_request(pin_id)
        resp = self._services.get_analog_io_service.call(req)

        if resp["status"] < RobotErrors.SUCCESS.value:
            return None

        return resp["value"]

    @property
    def custom_button_state(self):
        """
        Returns the io state client which can be used synchronously or asynchronously to obtain the io states.
        The button state client returns the custom button state.
        Only available on Ned2.

        Examples: ::

            # Get last value
            io.custom_button_state()
            io.custom_button_state.value

            # Subscribe a callback
            def button_callback(button_state):
                print button_state

            io.custom_button_state.subscribe(button_callback)
            io.custom_button_state.unsubscribe()

        :return: button state topic instance
        :rtype: NiryoTopic
        """
        return self.custom_button_state

    def is_custom_button_pressed(self):
        """
        Read the state of the custom button of the ned2.
        Only available on Ned2.

        :rtype: bool
        """
        return self.custom_button_state.value



from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.tool.enums import ToolID


class ToolTopics(object):

    def __init__(self, client):
        self.__client = client

        self.tool_id_topic = NiryoTopic(self.__client,
                                        '/niryo_robot_tools_commander/current_id',
                                        'std_msgs/Int32',
                                        tool_id_topic_conversion)


def tool_id_topic_conversion(msg):
    return ToolID(msg['data'])



#!/usr/bin/env python
import time
import unittest
import numpy as np
from threading import Event

from pyniryo2.objects import PoseObject
from pyniryo2.niryo_ros import NiryoRos
from pyniryo2.exceptions import RobotCommandException

from pyniryo2.arm.arm import Arm
from pyniryo2.arm.objects import JointStateObject, HardwareStatusObject
from pyniryo2.arm.enums import CalibrateMode, RobotAxis

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["test_hardware_status",
              "test_synchronous_calibration",
              #"test_calibration_callback",
              #"test_request_new_calibration_callback",
              # "test_learning_mode",
              "test_joints_state",
              "test_pose_state",
              "test_move_joints",
              "test_move_pose",
              "test_move_linear",
              "test_stop_move",
              "test_shift",
              "test_jog_joints",
              "test_jog_pose",
              "test_velocity",
              "test_kinematics"]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.arm = Arm(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()

    @staticmethod
    def assertAlmostEqualVector(a, b, decimal=1):
        np.testing.assert_almost_equal(a, b, decimal)


# noinspection PyTypeChecker
class TestHardwareStatus(BaseTest):

    def test_hardware_status(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsInstance(self.arm.hardware_status(), HardwareStatusObject)
        self.assertIsInstance(self.arm.hardware_status.value, HardwareStatusObject)

    def test_synchronous_calibration(self):
        self.assertIsNone(time.sleep(1))
        # Test Reset calibration
        self.assertIsNone(self.arm.reset_calibration())
        self.assertTrue(self.arm.need_calibration())

        # Test Auto calibration
        self.assertIsNone(self.arm.calibrate(CalibrateMode.AUTO))
        self.assertFalse(self.arm.need_calibration())
        self.assertIsNone(self.arm.reset_calibration())
        self.assertTrue(self.arm.need_calibration())
        self.assertIsNone(self.arm.calibrate_auto())
        self.assertFalse(self.arm.need_calibration())

        # Test Manual Calibration
        self.assertIsNone(self.arm.reset_calibration())
        self.assertTrue(self.arm.need_calibration())
        self.assertIsNone(self.arm.calibrate(CalibrateMode.MANUAL))
        self.assertFalse(self.arm.need_calibration())

    def test_calibration_callback(self):
        self.assertIsNone(time.sleep(1))
        calibration_event = Event()
        calibration_event.clear()

        def callback_calibration_end(response):
            self.assertIsNotNone(response)
            self.assertIsNone(time.sleep(1))
            calibration_event.set()

        self.assertIsNone(self.arm.reset_calibration())
        self.assertTrue(self.arm.need_calibration())

        self.assertIsNone(self.arm.calibrate_auto(callback=callback_calibration_end))
        self.assertTrue(calibration_event.wait(30))
        self.assertIsNone(time.sleep(1))
        self.assertFalse(self.arm.need_calibration())

    def test_request_new_calibration_callback(self):
        calibration_event = Event()
        calibration_event.clear()

        def callback_calibration_end(response):
            self.assertIsNotNone(response)
            self.assertIsNone(time.sleep(1))
            calibration_event.set()

        self.assertIsNone(self.arm.calibrate(CalibrateMode.MANUAL))
        self.assertFalse(self.arm.need_calibration())

        self.assertIsNone(self.arm.request_new_calibration(callback=callback_calibration_end))
        self.assertTrue(self.arm.need_calibration())
        self.assertTrue(calibration_event.wait(30))
        self.assertIsNone(time.sleep(1))
        self.assertFalse(self.arm.need_calibration())

    def test_learning_mode(self):
        self.assertIsNone(time.sleep(1))
        def setter_learning_mode(state):
            self.arm.learning_mode = state

        self.assertIsNone(self.arm.set_learning_mode(False))
        self.assertFalse(self.arm.learning_mode())

        self.assertIsNone(self.arm.set_learning_mode(True))
        self.assertTrue(self.arm.learning_mode())

        self.assertIsNone(setter_learning_mode(False))
        self.assertFalse(self.arm.get_learning_mode())

        self.assertIsNone(setter_learning_mode(True))
        self.assertTrue(self.arm.get_learning_mode())

        self.assertIsNone(time.sleep(2))

    def test_joints_state(self):
        self.assertIsNone(time.sleep(1))
        self.arm.learning_mode = False
        self.assertIsNone(time.sleep(1))
        self.assertAlmostEqualVector(self.arm.joints_state().position, self.arm.joints)
        self.assertAlmostEqualVector(self.arm.joints_state.value.position, self.arm.get_joints())
        self.assertIsInstance(self.arm.joints_state(), JointStateObject)
        #self.arm.learning_mode = True

    def test_pose_state(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsNotNone(self.arm.get_pose())
        self.assertIsNotNone(self.arm.get_pose().x)
        self.assertIsNotNone(self.arm.get_pose().y)
        self.assertIsNotNone(self.arm.get_pose().z)
        self.assertIsNotNone(self.arm.get_pose().roll)
        self.assertIsNotNone(self.arm.get_pose().pitch)
        self.assertIsNotNone(self.arm.get_pose().yaw)
        self.assertIsNotNone(self.arm.get_pose.value)
        self.assertIsNotNone(self.arm.pose)
        self.assertIsNotNone(self.arm.get_pose_quat())

    def test_move_joints(self):
        self.assertIsNone(time.sleep(1))
        def setter_joints(joints):
            self.arm.joints = joints

        self.arm.set_learning_mode(False)

        self.assertIsNone(self.arm.move_joints(6 * [0.2]))
        self.assertAlmostEqualVector(self.arm.joints, 6 * [0.2])

        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        self.assertIsNone(self.arm.move_joints(6 * [-0.2], end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertAlmostEqualVector(self.arm.joints, 6 * [-0.2])

        self.assertIsNone(setter_joints(6 * [0.2]))
        self.assertAlmostEqualVector(self.arm.get_joints(), 6 * [0.2])

        self.assertIsNone(self.arm.move_to_home_pose())
        self.assertAlmostEqualVector(self.arm.get_joints(), [0.0, 0.3, -1.3, 0.0, 0.0, 0.0])

        self.assertIsNone(self.arm.go_to_sleep())

    def test_move_pose(self):
        self.assertIsNone(time.sleep(1))
        def setter_pose(pose):
            self.arm.pose = pose

        self.assertIsNone(self.arm.move_pose(PoseObject(0.2, 0, 0.1, 0, 1.5, 0)))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0, 0.1, 0, 1.5, 0])

        self.assertIsNone(setter_pose([0.2, 0, 0.3, 0, 1.5, 0]))
        self.assertAlmostEqualVector(self.arm.pose.to_list(), [0.2, 0, 0.3, 0, 1.5, 0])

        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        self.assertIsNone(self.arm.move_pose([0.1, 0, 0.2, 0, 1.5, 0], end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertAlmostEqualVector(self.arm.pose.to_list(), [0.1, 0, 0.2, 0, 1.5, 0])

        with self.assertRaises(RobotCommandException):
            self.arm.move_pose([0.54, 0.964, 0.34, "a", "m", CalibrateMode.AUTO])

        #self.arm.set_learning_mode(True)

    def test_move_linear(self):
        self.assertIsNone(time.sleep(1))
        pose1 = [0.2, -0.1, 0.3, 0, 0, 0]
        self.assertIsNone(self.arm.move_pose(pose1))
        self.assertIsNone(time.sleep(0.5))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), pose1)

        pose2 = [0.2, 0., 0.4, 0, 0, 0]
        self.assertIsNone(self.arm.move_linear_pose(pose2))
        self.assertIsNone(time.sleep(0.5))
        self.assertAlmostEqualVector(self.arm.pose.to_list(), pose2)

        pose3 = [0.2, 0.1, 0.3, 0, 0, 0]
        self.assertIsNone(self.arm.move_linear_pose(PoseObject(*pose3)))
        self.assertIsNone(time.sleep(0.5))
        self.assertAlmostEqualVector(self.arm.pose.to_list(), pose3)

        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        pose4 = [0.2, -0.1, 0.3, 0, 0, 0]
        self.assertIsNone(self.arm.move_linear_pose(pose4, end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertAlmostEqualVector(self.arm.pose.to_list(), pose4)

        with self.assertRaises(RobotCommandException):
            self.arm.move_linear_pose([0.54, 0.964, "m", CalibrateMode.AUTO])

        #self.arm.set_learning_mode(True)

    def test_stop_move(self):
        self.assertIsNone(time.sleep(1))
        joints = [0., 0., 0., 0., 0., 0.]
        self.assertIsNone(self.arm.move_joints(joints))
        self.assertAlmostEqualVector(self.arm.joints, joints)

        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        joints = [0.0, 0.3, -1.3, 0.0, 0.0, 0.0]
        self.assertIsNone(self.arm.move_joints(joints, end_move_callback))
        self.assertIsNone(time.sleep(1))
        self.assertIsNone(self.arm.stop_move())
        self.assertTrue(end_move_event.wait(10))

        with self.assertRaises(AssertionError):
            self.assertAlmostEqualVector(self.arm.joints, joints)

        self.assertIsNone(time.sleep(1))
        self.assertIsNone(self.arm.go_to_sleep())

    def test_shift(self):
        self.assertIsNone(time.sleep(3))
        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        self.assertIsNone(self.arm.move_pose([0.2, 0.1, 0.1, 0, 1., 0]))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0.1, 0.1, 0, 1., 0])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.X, 0.1))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.3, 0.1, 0.1, 0, 1., 0])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.Y, -0.1))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.3, 0., 0.1, 0, 1., 0])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.Z, 0.1, end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.3, 0., 0.2, 0, 1., 0])

        self.assertIsNone(self.arm.move_pose([0.2, 0., 0.3, 0, 0, 0]))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0., 0.3, 0, 0, 0])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.PITCH, 1.))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0., 0.3, 0, 1., 0])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.ROLL, 0.5))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0., 0.3, 0.5, 1., 0.])

        self.assertIsNone(self.arm.shift_pose(RobotAxis.YAW, -.5))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.2, 0., 0.3, 0.5, 1., -.5])

        self.assertIsNone(self.arm.go_to_sleep())

    def test_jog_joints(self):
        self.assertIsNone(time.sleep(3))
        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        self.assertIsNone(self.arm.move_joints(6 * [0.0]))
        self.assertAlmostEqualVector(self.arm.get_joints(), 6 * [0.0])

        self.assertIsNone(self.arm.set_jog_control(True))

        self.assertIsNone(self.arm.jog_joints([-0.2, 0.0, 0.2, 0.2, 0, 0]))
        self.assertIsNone(time.sleep(2))
        self.assertAlmostEqualVector(self.arm.get_joints(), [-0.2, 0.0, 0.2, 0.2, 0, 0])

        self.assertIsNone(self.arm.jog_joints([0.2, 0.0, -0.2, -0.2, 0, 0], end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertIsNone(time.sleep(2))
        self.assertAlmostEqualVector(self.arm.get_joints(), 6 * [0.0])

        # Check Exceptions
        with self.assertRaises(RobotCommandException):
            self.arm.jog_joints([0.1, 0.0, -0.1], 0.05)

        self.assertIsNone(self.arm.set_jog_control(False))

    def test_jog_pose(self):
        self.assertIsNone(time.sleep(3))
        end_move_event = Event()
        end_move_event.clear()

        def end_move_callback(_):
            end_move_event.set()

        self.assertIsNone(self.arm.move_pose([0.3, 0.0, 0.3, 0.0, 0.0, 0.0]))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), [0.3, 0.0, 0.3, 0.0, 0.0, 0.0])

        self.assertIsNone(self.arm.jog_pose([-0.01, 0.01, 0.01, 0., 0., 0.]))
        self.assertIsNone(time.sleep(0.5))
        self.assertAlmostEqualVector(self.arm.get_pose().to_list()[:3], [0.29, 0.01, 0.31])

        pose1 = self.arm.get_pose().to_list()
        self.assertIsNone(self.arm.jog_pose([0., 0., 0., 0.1, 0.0, 0.0], end_move_callback))
        self.assertTrue(end_move_event.wait(10))
        self.assertIsNone(time.sleep(0.5))
        pose1[3] += 0.1
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), pose1)

        pose2 = self.arm.get_pose().to_list()
        self.assertIsNone(self.arm.jog_pose([0., 0., 0., 0., -0.1, 0.]))
        self.assertIsNone(time.sleep(0.5))
        pose2[4] -= 0.1
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), pose2)

        pose3 = self.arm.get_pose().to_list()
        self.assertIsNone(self.arm.jog_pose([0., 0., 0., 0., 0., 0.1]))
        self.assertIsNone(time.sleep(0.5))
        pose3[5] += 0.1
        self.assertAlmostEqualVector(self.arm.get_pose().to_list(), pose3)

        # Check Exceptions
        with self.assertRaises(RobotCommandException):
            self.arm.jog_pose([0.1, 0.0, -0.1], 0.05)

        self.assertIsNone(self.arm.set_jog_control(False))

    def test_velocity(self):
        self.assertIsNone(time.sleep(1))
        self.assertIsNone(self.arm.set_arm_max_velocity(100))
        self.assertEqual(self.arm.get_arm_max_velocity(), 100)

        self.velocity = None

        def velocity_callback(msg):
            self.velocity = msg

        self.assertIsNone(self.arm.get_arm_max_velocity.subscribe(velocity_callback))
        self.assertIsNone(self.arm.set_arm_max_velocity(50))
        self.assertIsNone(time.sleep(1))
        self.assertEqual(self.velocity, 50)

        with self.assertRaises(RobotCommandException):
            self.arm.set_arm_max_velocity(101)

        with self.assertRaises(RobotCommandException):
            self.arm.set_arm_max_velocity(-1)

        with self.assertRaises(RobotCommandException):
            self.arm.set_arm_max_velocity(0)

        self.assertIsNone(time.sleep(0.5))
        self.assertIsNone(self.arm.set_arm_max_velocity(100))
        self.assertIsNone(time.sleep(0.5))
        self.assertEqual(self.arm.get_arm_max_velocity(), 100)


    def test_kinematics(self):
        self.assertIsNone(time.sleep(1))
        # Forward Kinematics
        joints_target = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        pose_target = self.arm.forward_kinematics(joints_target)
        self.assertIsNone(self.arm.move_pose(pose_target))
        self.assertAlmostEqualVector(pose_target.to_list(), self.arm.get_pose().to_list())
        joints_reached = self.arm.get_joints()
        self.assertAlmostEqualVector(joints_target, joints_reached)

        # Inverse Kinematics
        pose_target = [0.2, 0.0, 0.3, 0.0, 0.0, 0.0]
        joints_target = self.arm.inverse_kinematics(pose_target)
        self.assertIsNone(self.arm.move_joints(joints_target))
        pose_reached = self.arm.get_pose()
        self.assertAlmostEqualVector(pose_target, pose_reached.to_list())

        #self.assertIsNone(self.arm.set_learning_mode(True))


def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestHardwareStatus(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



"""
PyNiryo2 utils
-----------------
A collection of utilities to convert data from dict to list and vice versa
"""


def pose_dict_to_list(pose_dict):
    return [pose_dict["position"][axis] for axis in ["x", "y", "z"]] + \
           [pose_dict["rpy"][axis] for axis in ["roll", "pitch", "yaw"]]


def pose_list_to_dict(pose_list):
    return {"position": dict(zip(["x", "y", "z"], pose_list[:3])),
            "rpy": dict(zip(["roll", "pitch", "yaw"], pose_list[3:]))}


def pose_quat_dict_to_list(pose_dict):
    return [pose_dict["position"][axis] for axis in ["x", "y", "z"]] + \
           [pose_dict["orientation"][axis] for axis in ["x", "y", "z", "w"]]


def pose_quat_list_to_dict(pose_list):
    return {"position": dict(zip(["x", "y", "z"], pose_list[:3])),
            "orientation": dict(zip(["x", "y", "z", "w"], pose_list[3:]))}


def point_dict_to_list(pose_dict):
    return [pose_dict[axis] for axis in ["x", "y", "z"]]


def point_list_to_dict(pose_list):
    return dict(zip(["x", "y", "z"], pose_list[:3]))



#!/usr/bin/env python
import time
import unittest
from threading import Event

from pyniryo2.exceptions import RobotCommandException
from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.niryo_ros import NiryoRos

from pyniryo2.io.enums import PinID, PinMode, PinState
from pyniryo2.io.objects import DigitalPinObject, AnalogPinObject
from pyniryo2.io.io import IO

robot_ip_address = "127.0.0.1"
port = 9090

test_order = ["get_digital_io_sate",
              "test_pin_mode",
              "test_set_digital_pin_state",
              ]


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)
        cls.client.run()
        cls.io = IO(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()


# noinspection PyTypeChecker
class TestIO(BaseTest):
    def get_digital_io_sate(self):
        for pin in [PinID.GPIO_1A, PinID.GPIO_1B, PinID.GPIO_2C, PinID.SW_1]:
            self.assertIsInstance(self.io.get_digital_io_states, NiryoTopic)
            self.assertIsInstance(self.io.get_digital_io_states(), list)
            self.assertIsInstance(self.io.get_digital_io_states()[0], DigitalPinObject)

            self.assertIsInstance(self.io.get_digital_io_state(pin), DigitalPinObject)
            self.assertIsInstance(self.io.digital_read(pin), bool)

            io_state = self.io.get_digital_io_state(pin)
            self.assertIsInstance(io_state.pin_id, PinID)
            self.assertIsInstance(io_state.name, str)
            self.assertIsInstance(io_state.mode, PinMode)
            self.assertIsInstance(io_state.value, bool)
            self.assertEqual(io_state.value, self.io.digital_read(pin))

            io_state_event = Event()
            io_state_event.clear()

            def io_state_callback(io_state_list):
                self.assertIsInstance(io_state_list, list)
                self.assertIsInstance(io_state_list[0], DigitalPinObject)
                io_state_event.set()

            self.assertIsNone(self.io.get_digital_io_states.subscribe(io_state_callback))
            self.assertTrue(io_state_event.wait(10))
            self.assertIsNone(self.io.get_digital_io_states.unsubscribe())

    def test_pin_mode(self):
        for pin_id in [PinID.GPIO_1A, PinID.GPIO_1B, PinID.GPIO_2C]:
            self.assertIsNone(self.io.set_pin_mode(pin_id, PinMode.OUTPUT))
            self.assertEqual(self.io.get_digital_io_state(pin_id), PinMode.OUTPUT)

            self.assertIsNone(self.io.set_pin_mode(pin_id, PinMode.INPUT))
            self.assertEqual(self.io.get_digital_io_state(pin_id), PinMode.INPUT)

        with self.assertRaises(RobotCommandException):
            self.io.set_pin_mode(PinID.SW_1, PinMode.INPUT)

    def test_set_digital_pin_state(self):
        # with self.assertRaises(RobotCommandException):
        #    self.io.digital_write(PinID.DI3, PinState.HIGH)

        for pin_id in [PinID.GPIO_1A, PinID.SW_1]:  # , PinID.DO4]:
            if not pin_id.value.startswith("SW"):
                self.assertIsNone(self.io.set_pin_mode(pin_id, PinMode.OUTPUT))
                self.assertEqual(self.io.get_digital_io_state(pin_id), PinMode.OUTPUT)

            self.assertIsNone(self.io.digital_write(pin_id, PinState.HIGH))
            time.sleep(0.1)
            self.assertEqual(self.io.get_digital_io_state(pin_id).value, True)
            self.assertEqual(self.io.digital_read(pin_id), True)
            self.assertIsNone(self.io.digital_write(pin_id, PinState.LOW))
            time.sleep(0.1)
            self.assertEqual(self.io.get_digital_io_state(pin_id).value, False)
            self.assertEqual(self.io.digital_read(pin_id), False)

            self.assertIsNone(self.io.digital_write(pin_id.value, True))
            time.sleep(0.1)
            self.assertEqual(self.io.get_digital_io_state(pin_id).value, True)
            self.assertEqual(self.io.digital_read(pin_id), True)
            self.assertIsNone(self.io.digital_write(pin_id.value, False))
            time.sleep(0.1)
            self.assertEqual(self.io.get_digital_io_state(pin_id).value, False)
            self.assertEqual(self.io.digital_read(pin_id), False)

        with self.assertRaises(RobotCommandException):
            self.io.digital_write(1, PinState.LOW)

        with self.assertRaises(RobotCommandException):
            self.io.digital_write(PinID.DO1, 0)



def suite():
    suite = unittest.TestSuite()
    for function_name in test_order:
        suite.addTest(TestIO(function_name))
    return suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



import roslibpy
from pyniryo2.tool.enums import ToolID
from pyniryo2.utils import pose_list_to_dict


class ToolServices(object):

    def __init__(self, client):
        self.__client = client

        self.update_tool_service = roslibpy.Service(self.__client,
                                                    '/niryo_robot_tools_commander/update_tool',
                                                    'niryo_robot_msgs/Trigger')

        self.equip_electromagnet_service = roslibpy.Service(self.__client,
                                                            '/niryo_robot_tools_commander/equip_electromagnet',
                                                            'niryo_robot_msgs/SetInt')

        self.enable_tcp_service = roslibpy.Service(self.__client,
                                                   '/niryo_robot_tools_commander/enable_tcp',
                                                   'niryo_robot_msgs/SetBool')

        self.set_tcp_service = roslibpy.Service(self.__client,
                                                '/niryo_robot_tools_commander/set_tcp',
                                                'niryo_robot_tools_commander/SetTCP')

        self.reset_tcp_service = roslibpy.Service(self.__client,
                                                  '/niryo_robot_tools_commander/reset_tcp',
                                                  'niryo_robot_msgs/Trigger')

    @staticmethod
    def get_trigger_request():
        return roslibpy.ServiceRequest()

    @staticmethod
    def equip_electromagnet_service_request(id_=ToolID.ELECTROMAGNET_1):
        return roslibpy.ServiceRequest({"value": id_.value})

    @staticmethod
    def enable_tcp_service_request(enable):
        return roslibpy.ServiceRequest({"value": enable})

    @staticmethod
    def set_tcp_service_request(pose):
        return roslibpy.ServiceRequest(pose_list_to_dict(pose))



import roslibpy

from .enums import AnimationMode


class LedRingServices(object):

    def __init__(self, client):
        self.__client = client

        self.set_led_ring_animation_service = roslibpy.Service(self.__client,
                                                               '/niryo_robot_led_ring/set_user_animation',
                                                               'niryo_robot_led_ring/LedUser')

        self.set_led_ring_led_color_service = roslibpy.Service(self.__client,
                                                               '/niryo_robot_led_ring/set_led_color',
                                                               'niryo_robot_led_ring/SetLedColor')

    def set_led_ring_request(self, animation_nb, color_list=None, period=0, iterations=0, wait=False):
        animation_mode = {'animation': animation_nb.value if isinstance(animation_nb, AnimationMode) else animation_nb}

        if not isinstance(color_list, list) or len(color_list) < 1:
            color_list_rgb = []
        elif not isinstance(color_list[0], list):
            color_list_rgb = [self.color_to_color_rgba(color) for color in [color_list]]
        else:
            color_list_rgb = [self.color_to_color_rgba(color) for color in color_list]

        return roslibpy.ServiceRequest({"animation_mode": animation_mode,
                                        "colors": color_list_rgb,
                                        "period": period,
                                        "iterations": iterations,
                                        "wait_end": wait})

    def set_led_ring_color_request(self, led_id, color):
        color_rgba = [] if not isinstance(color, list) else color[:]
        return roslibpy.ServiceRequest({"led_id": led_id,
                                        "color": self.color_to_color_rgba(color_rgba)})

    @staticmethod
    def color_to_color_rgba(color):
        color_rgb = color if len(color) >= 3 else color[:] + (3 - len(color)) * [0]
        return {'r': color_rgb[0], 'g': color_rgb[1], 'b': color_rgb[2], 'a': 0}



# Python libraries
import roslibpy
import socket
import time
from threading import Thread

from pyniryo2.niryo_topic import NiryoTopic


class NiryoRosTimeoutException(Exception):
    pass


class NiryoRos(roslibpy.Ros):
    def __init__(self, ip_address="127.0.0.1", port=9090):
        """
        Connect to your computer to ros: ::

            ros_instance = NiryoRos("127.0.0.1") # Simulation

            ros_instance = NiryoRos("10.10.10.10") # Hotspot

            ros_instance = NiryoRos("169.254.200.201") # Ethernet

        Based on the roslibpy ROS client: https://roslibpy.readthedocs.io/en/latest/reference/index.html#roslibpy.Ros

        :param ip_address: ip of the ros master
        :type ip_address: str
        :param port: usually 9090
        :type port: int
        """
        self.port = port
        self.ip_address = ip_address
        super(NiryoRos, self).__init__(host=ip_address, port=port)

        self.should_run = False
        self.__pyniryo2_ip = socket.gethostbyname(socket.gethostname())

        self.__ping_loop_thread = Thread(target=self.__ping_loop)
        self.__pyniryo_ping_publisher = NiryoTopic(self, '/ping_pyniryo', 'std_msgs/Bool')

        self.run(timeout=5)
        self.__ping_loop_thread.start()
        self._hardware_version = self.get_param("/niryo_robot/hardware_version")
        time.sleep(0.5)

    def close(self):
        self.should_run = False
        super(NiryoRos, self).close()

    def terminate(self):
        self.should_run = False
        super(NiryoRos, self).terminate()

    def __ping_loop(self):
        self.should_run = True
        while self.is_connected and self.should_run:
            self.__pyniryo_ping_publisher.publish(roslibpy.Message({'data': True}))
            time.sleep(0.1)

    @property
    def hardware_version(self):
        """
        Get the hardware version of the robot (one, ned, ned2)

        :return: The hardware version of the robot (one, ned, ned2)
        :rtype: str
        """
        return self._hardware_version



#!/usr/bin/env python

import numpy as np
import sys
import unittest
import roslibpy
from threading import Event

from pyniryo2.niryo_topic import NiryoTopic
from pyniryo2.exceptions import TopicException
from pyniryo2.niryo_ros import NiryoRos

robot_ip_address = "127.0.0.1"
port = 9090


class BaseTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.client = NiryoRos(ip_address=robot_ip_address, port=port)

    @classmethod
    def tearDownClass(cls):
        cls.client.close()

    @staticmethod
    def assertAlmostEqualVector(a, b, decimal=1):
        np.testing.assert_almost_equal(a, b, decimal)


# noinspection PyTypeChecker
class TestTopic(BaseTest):

    def setUp(self):
        self.topic_name = '/joint_states'
        self.topic_type = 'sensor_msgs/JointState'

    def test_subscribe_unsubscribe(self):
        self.topic_message = None
        self.message_event = Event()
        self.message_event.clear()

        def callback(message):
            self.topic_message = message
            self.message_event.set()

        topic = NiryoTopic(self.client, self.topic_name, self.topic_type)

        self.assertFalse(topic.is_subscribed)

        # Try a bad subscription
        with self.assertRaises(TopicException):
            topic.subscribe(None)

        # Subscribe
        self.assertIsNone(topic.subscribe(callback))
        self.assertTrue(topic.is_subscribed)

        # Wait a message
        self.assertTrue(self.message_event.wait(timeout=0.5))
        self.assertIsNotNone(self.topic_message)

        # Unsubscibe
        self.assertIsNone(topic.unsubscribe())
        self.assertFalse(topic.is_subscribed)

    def test_type_conversion(self):
        def conversion_get_joint_name(msg):
            return str(msg["name"][0])

        topic = NiryoTopic(self.client, self.topic_name, self.topic_type, conversion_function=conversion_get_joint_name)
        self.assertIsInstance(topic(), str)

        def conversion_get_joint_value(msg):
            return float(msg["position"][0])

        topic = NiryoTopic(self.client, self.topic_name, self.topic_type,
                           conversion_function=conversion_get_joint_value)
        self.assertIsInstance(topic(), float)

    def test_synchronous_access(self):
        self.assertTrue(self.client.is_connected)

        topic = NiryoTopic(self.client, self.topic_name, self.topic_type)
        self.assertFalse(topic.is_subscribed)

        # Wait a message
        message = topic()
        self.assertIsNotNone(message)
        self.assertNotEqual(message, topic())

        self.assertFalse(topic.is_subscribed)

    def test_synchronous_with_asynchronous_access(self):
        self.topic_message = None
        self.message_event = Event()
        self.message_event.clear()

        def callback(message):
            self.assertEqual(topic(), message)
            self.topic_message = message
            self.message_event.set()

        topic = NiryoTopic(self.client, self.topic_name, self.topic_type)

        # Synchronous access
        self.assertIsNotNone(topic())
        self.assertFalse(topic.is_subscribed)

        # Subscribe
        self.assertIsNone(topic.subscribe(callback))
        self.assertTrue(topic.is_subscribed)

        # Wait a message
        self.assertTrue(self.message_event.wait(timeout=0.5))
        self.assertIsNotNone(self.topic_message)

        # Unsubscibe
        self.assertIsNone(topic.unsubscribe())
        self.assertFalse(topic.is_subscribed)


if __name__ == '__main__':
    unittest.main()



# !/usr/bin/env python

# Imports
from gc import callbacks
from pyniryo2 import *
from threading import Event

from trajectory_msgs.msg import JointTrajectory, JointTrajectoryPoint
from std_msgs.msg import Int32, Bool, Header

robot_ip = "192.168.1.115"
robot_ip_address_local = "127.0.0.1"

# Events
update_tool_event = Event()
update_tool_event.clear()

calibrated_event = Event()
calibrated_event.clear()

save_trajectory_event = Event()
save_trajectory_event.clear()

executed_trajectory_event = Event()
executed_trajectory_event.clear()

get_trajectory_list_event = Event()
get_trajectory_list_event.clear()

update_trajectory_event = Event()
update_trajectory_event.clear()

delete_trajectory_event = Event()
delete_trajectory_event.clear()

# Callbacks
def update_tool_success_callback(result):
    update_tool_event.set()
    print ('Update Tool: ', result['message'])
    print ("")

def update_tool_error_callback(result):
    print ('Update Tool: ', result['message'])
    print ("")

def calibrate_success_callback(result):
    calibrated_event.set()
    print ("")
    print ('Calibration : ', result["message"])
    print ("")

def calibrate_error_callback(result):
    print ('Calibration : ', result["message"])
    print ("")

def execute_registered_callback(result):
    print("Trajectory Callback : ", result["message"])
    print ("")
    executed_trajectory_event.set()

def get_trajectory_callback(result):
    print("Get Trajectory Callback : ", result["trajectory"])
    print ("")

def get_trajectory_list_callback(result):
    print("Trajectory List : ", result["name_list"])
    print("Description List : ", result["description_list"])
    print ("")
    get_trajectory_list_event.set()

def save_trajectory_callback(result):
    print("Save Trajectory Callback : ", result["message"])
    print ("")
    save_trajectory_event.set()

def delete_trajectory_callback(result):
    print("Delete Trajectory : ", result["message"])
    print ("")
    delete_trajectory_event.set()

def update_callback(result):
    print("Update Trajectory : ", result["message"])
    print("")
    update_tool_event.set()

def action_function(robot):

    """
    Don't put niryo_robot in parameter, it will take the pyniryo2 package
    add your function here
    """
    robot.trajectories.get_saved_trajectory("last_executed_trajectory")

    robot.trajectories.execute_registered_trajectory("last_executed_trajectory", callback=execute_registered_callback)
    executed_trajectory_event.wait(20)
    executed_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

    joints_list = [[-0.493, -0.32, -0.505, -0.814, -0.282, 0],
              [0.834, -0.319, -0.466, 0.822, -0.275, 0],
              [1.037, -0.081, 0.248, 1.259, -0.276, 0]]

    points_list = [JointTrajectoryPoint(positions=joints) for joints in joints_list]

    trajectory = JointTrajectory(header = Header(), joint_names =["joint_1", "joint_2", "joint_3",
             "joint_4", "joint_5", "joint_6"], points = points_list)

    robot.trajectories.save_trajectory(trajectory, "test_trajectory", "ceci est un test", callback=save_trajectory_callback)
    save_trajectory_event.wait(10)
    save_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

    robot.trajectories.delete_trajectory("test_trajectory", callback=delete_trajectory_callback)
    delete_trajectory_event.wait(10)
    delete_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

    print ('Execute Trajectory with the arm')
    robot.wait(10)
    response = input("Done ? [Press Enter]")
    print ("")

    robot.trajectories.save_last_learned_trajectory("test_py2", "coucou", callback=update_callback)
    update_trajectory_event.wait(10)
    update_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

    robot.trajectories.execute_registered_trajectory("test_py2", callback=execute_registered_callback)
    executed_trajectory_event.wait(50)
    executed_trajectory_event.clear()

    robot.trajectories.update_trajectory_infos("test_py2", "traj_changes_name", "ah que coucou", callback=update_callback)
    update_trajectory_event.wait(10)
    update_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

    robot.trajectories.delete_trajectory("traj_changes_name", callback=delete_trajectory_callback)
    delete_trajectory_event.wait(10)
    delete_trajectory_event.clear()

    robot.trajectories.get_saved_trajectory_list(callback=get_trajectory_list_callback)
    get_trajectory_list_event.wait(10)
    get_trajectory_list_event.clear()

if __name__ == "__main__":

    # Connect to robot
    robot = NiryoRobot(robot_ip)

    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto(callback=calibrate_success_callback, errback=calibrate_error_callback)
    calibrated_event.wait(20)
    if not calibrated_event.is_set():
        quit

    robot.tool.update_tool(callback=update_tool_success_callback, errback=update_tool_error_callback)
    update_tool_event.wait()

    action_function(robot)

    # Releasing connection
    robot.arm.go_to_sleep()
    robot.end()



from pyniryo2.niryo_topic import NiryoTopic


class SoundTopics(object):

    def __init__(self, client):
        self.__client = client

        self.sound_database_topic = NiryoTopic(self.__client,
                                               '/niryo_robot_sound/sound_database',
                                               'niryo_robot_sound/SoundList',
                                               sound_database_conversion)

        self.current_sound_topic = NiryoTopic(self.__client,
                                              '/niryo_robot_sound/sound',
                                              'std_msgs/String',
                                              current_sound_conversion)

        self.volume_topic = NiryoTopic(self.__client,
                                       '/niryo_robot_sound/volume',
                                       'std_msgs/UInt8',
                                       volume_sound_conversion)


def sound_database_conversion(msg):
    return {str(sound["name"]): sound["duration"] for sound in msg["sounds"]}


def current_sound_conversion(msg):
    return str(msg["data"])


def volume_sound_conversion(msg):
    return int(msg["data"])



# !/usr/bin/env python3

# Imports
from pyniryo2 import *
from threading import Event

robot_ip, default_workspace_name = "192.168.1.124", "test_ws-2"

# Pointer Offset
pointer_offset = 0.05

# Events
update_tool_event = Event()
update_tool_event.clear()

calibrated_event = Event()
calibrated_event.clear()

move_pose_event = Event()
move_pose_event.clear()

close_event = Event()
close_event.clear()

# Poses - Modify According to your setup
observation_pose = PoseObject(x=0.220,y=0.0, z=0.376, roll=-1.51, pitch=1.526, yaw=-1.395)

place_pose = PoseObject(x=0.0, y=0.268, z=0.17, roll=-0.974, pitch=1.5, yaw=0.72)

# Callbacks
def update_tool_success_callback(result):
    update_tool_event.set()
    print 'Update Tool: ', result['message']
    print ("")

def update_tool_error_callback(result):
    print 'Update Tool: ', result['message']
    print ("")

def calibrate_success_callback(result):
    calibrated_event.set()
    print 'Calibrate Callback: ', result["message"]
    print ("")

def calibrate_error_callback(result):
    print 'Calibrate Callback: ', result["message"]
    print ("")

def close_gripper_callback(result):
    close_event.set()
    print 'Close Callback: ', result["message"]
    print ("")

def move_pose_callback(result):
    move_pose_event.set()
    print 'Close Callback: ', result["message"]
    print ("")

def check_ws(robot, workspace_name):
    ws_list = robot.vision.get_workspace_list()
    if len(ws_list) !=0 and workspace_name in ws_list:
        print ("Workspace Found !")
        print ("")
    else:
        print("Workspace Not Found ! Please define a Workspace name !")
        print ("")
        workspace_name = workspace_definition(robot)
    return workspace_name

def workspace_name_definition():
    ws_validation = False
    while not ws_validation:
        workspace_response = raw_input("Workspace Name ? ")
        print ("")
        response = raw_input("Validate ? [y/n] : ")
        while response not in ["y","n"]:
            print ("Please enter <y> or <n>")
            print ("")
            response = raw_input("Validate ? [y/n] : ")
            print ("")
        if response == "y":
            workspace_name = workspace_response
            ws_validation = True
        elif response == "n":
            print ("Ok ! Let's try again !")
    return workspace_name

def workspace_definition(robot):
    print ("")
    print ("Please set Pointer on the robot")
    response = raw_input("Done ? [Press Enter] ")
    print ("")
    robot.wait(1)
    workspace_name = workspace_name_definition()
    workspace_poses = []
    for i in range(4):
        validation = False
        while not validation:
            ws_list = robot.vision.get_workspace_list()
            if len(ws_list) !=0 and workspace_name in ws_list:
                robot.vision.delete_workspace(workspace_name)
            print ("")
            print ("Place Pointer on workspace pose number "+str(i+1))
            response = raw_input("Done ? [Press Enter]")
            x, y, z, roll, pitch, yaw = robot.arm.get_pose().to_list()
            print ("")
            print ("Pose "+str(i+1)+": "+str(x)+" "+str(y)+" "+str(z)+" "+str(roll)+" "+str(pitch)+" "+str(yaw))
            print ("")
            response = raw_input("Validate ? [y/n] : ")
            print ("")
            while response not in ['y','n']:
                print ("Please enter <y> or <n>")
                print ("")
                response = raw_input("Validate ? [y/n] : ")
                print ("")
            if response == 'y':
                print ("Point Validated")
                print ("")
                Pose = PoseObject(x, y, z, roll, pitch, yaw)
                workspace_poses.append(Pose)
                validation = True
            elif response == 'n':
                print ("Ok ! Let's try again !")
                print ("")
    print("Recap: ")
    print (workspace_name)
    print("Pose 1", workspace_poses[0])
    print("Pose 2", workspace_poses[1])
    print("Pose 3", workspace_poses[2])
    print("Pose 4", workspace_poses[3])
    print ("")
    robot.arm.move_to_home_pose()
    robot.vision.save_workspace_from_robot_poses(workspace_name,workspace_poses[0], workspace_poses[1], workspace_poses[2], workspace_poses[3])
    return workspace_name
    

def vision_pick_n_place(robot, workspace_name):
    """
    Simple pick and place with vision: Example 2

    :type niyro_robot: NiryoRobot
    :rtype: None
    """
    # Loop
    try_without_success = 0
    while try_without_success < 5:
        # Moving to observation pose
        robot.arm.move_pose(observation_pose, callback=move_pose_callback)
        move_pose_event.wait()
        robot.wait(5)

        # Trying to pick target using camera
        ret = robot.vision.vision_pick(workspace_name,
                                                 height_offset=-0.02,
                                                 shape=ObjectShape.ANY,
                                                 color=ObjectColor.BLUE)
        obj_found, shape_ret, color_ret = ret
        if not obj_found:
            try_without_success += 1
            continue
        # Vision pick has succeed which means that Ned should have already catch the object !

        # Everything is good, so we going to place the object
        robot.pick_place.place_from_pose(place_pose)
        break

if __name__ == "__main__":
    
    # Connect to robot
    robot = NiryoRobot(robot_ip)
    
    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto(callback=calibrate_success_callback, errback=calibrate_error_callback)
    calibrated_event.wait(20)
    if not calibrated_event.is_set():
        quit

    workspace_def_response = raw_input("Do you want to define a workspace ? [y/n] : ")
    print ("")
    while workspace_def_response not in ['y','n']:
        print ("Please enter <y> or <n>")
        print ("")
        workspace_def_response = raw_input("Validate ? [y/n] : ")
        print ("")
    if workspace_def_response == 'y':
        print("Let's define the vision workspace !")
        workspace_name = workspace_definition(robot)
    elif workspace_def_response == 'n':
        print ("Ok ! Let's check your default workspace")
        workspace_name = check_ws(robot, default_workspace_name)

    robot.wait(1)
    print ("Please set Tool on the robot")
    print ("")
    response = raw_input("Done ? [Press Enter] ")

    robot.arm.move_to_home_pose()

    robot.tool.update_tool(callback=update_tool_success_callback, errback=update_tool_error_callback)
    update_tool_event.wait()

    vision_pick_n_place(robot, workspace_name)

    robot.tool.close_gripper(callback=close_gripper_callback)
    close_event.wait()
    # Releasing connection
    robot.arm.go_to_sleep()
    robot.end()





# Communication imports
from pyniryo2.robot_commander import RobotCommander
from pyniryo2.objects import PoseObject

from .services import SavedPosesServices


class SavedPoses(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        super(SavedPoses, self).__init__(client)

        self._services = SavedPosesServices(self._client)

    def get_pose_saved(self, pose_name):
        """
        Get pose saved in from Ned's memory

        Examples: ::

            pose = saved_poses.get_pose_saved("pose1")

        :param pose_name: Pose name in robot's memory
        :type pose_name: str
        :return: Pose associated to pose_name
        :rtype: PoseObject
        """
        self._check_type(pose_name, str)

        req = self._services.get_saved_pose_request(pose_name)
        resp = self._services.get_pose_service.call(req)

        self._check_result_status(resp)
        return self._services.get_saved_pose_response_to_pose(resp)

    def save_pose(self, pose_name, *args):
        """
        Save pose (x, y, z, roll, pitch, yaw) in robot's memory

        Examples: ::

            saved_poses.save_pose("pose1", 0.3, 0.0, 0.3, 0.0, 1.57, 0.0)
            saved_poses.save_pose("pose1", [0.3, 0.0, 0.3, 0.0, 1.57, 0.0])
            saved_poses.save_pose("pose1", PoseObject(0.3, 0.0, 0.3, 0.0, 1.57, 0.0))

        :type pose_name: str
        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a PoseObject
        :type args: Union[list[float], tuple[float], PoseObject]
        :rtype: None
        """
        self._check_type(pose_name, str)

        req = self._services.save_pose_request(pose_name, self._args_pose_to_list(*args))
        resp = self._services.manage_pose_service.call(req)

        self._check_result_status(resp)

    def delete_pose(self, pose_name):
        """
        Delete pose from robot's memory

        Examples: ::

            if "pose1" in saved_poses.get_saved_pose_list():
                saved_poses.delete_pose("pose1")

        :type pose_name: str
        :rtype: None
        """
        self._check_type(pose_name, str)

        req = self._services.delete_pose_request(pose_name)
        resp = self._services.manage_pose_service.call(req)

        self._check_result_status(resp)

    def get_saved_pose_list(self):
        """
        Get list of poses' name saved in robot memory

        Examples: ::

            >> print(saved_poses.get_saved_pose_list())
            ["pose1", "pose2", "pose3"]

        :rtype: list[str]
        """
        req = self._services.get_pose_list_request()
        resp = self._services.get_pose_list_service.call(req)
        return self._services.get_saved_pose_list_response_to_list(resp)



from pyniryo2 import NiryoRobot  # Niryo's Python Package
import Tkinter as Tk

simulation_mode = True

DEFAULT_XYZ_STEP = 0.01
DEFAULT_RPY_STEP = 0.1
DEFAULT_JOINT_STEP = 0.1

# Set robot address
robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"
robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi


class TkinterController(Tk.Tk):
    def __init__(self, niyro_robot):
        """

        :param niyro_robot: Client
        :type niyro_robot: NiryoRobot
        """
        Tk.Tk.__init__(self)
        self.niyro_robot = niyro_robot
        self.niyro_robot.tool.update_tool()

        self.xyz_step = DEFAULT_XYZ_STEP
        self.rpy_step = DEFAULT_RPY_STEP
        self.joint_step = DEFAULT_JOINT_STEP

        self.button_x_plus = Tk.Button(self, text="X+", command=self.x_plus).grid(row=0, column=0)
        self.button_x_minus = Tk.Button(self, text="X-", command=self.x_minus).grid(row=0, column=1)

        self.button_y_plus = Tk.Button(self, text="Y+", command=self.y_plus).grid(row=1, column=0)
        self.button_y_minus = Tk.Button(self, text="Y-", command=self.y_minus).grid(row=1, column=1)

        self.button_z_plus = Tk.Button(self, text="Z+", command=self.z_plus).grid(row=2, column=0)
        self.button_z_minus = Tk.Button(self, text="Z-", command=self.z_minus).grid(row=2, column=1)

        self.button_roll_plus = Tk.Button(self, text="roll+", command=self.roll_plus).grid(row=3, column=0)
        self.button_roll_minus = Tk.Button(self, text="roll-", command=self.roll_minus).grid(row=3, column=1)

        self.button_pitch_plus = Tk.Button(self, text="pitch+", command=self.pitch_plus).grid(row=4, column=0)
        self.button_pitch_minus = Tk.Button(self, text="pitch-", command=self.pitch_minus).grid(row=4, column=1)

        self.button_yaw_plus = Tk.Button(self, text="yaw+", command=self.yaw_plus).grid(row=5, column=0)
        self.button_yaw_minus = Tk.Button(self, text="yaw-", command=self.yaw_minus).grid(row=5, column=1)

        self.button_j1_plus = Tk.Button(self, text="j1+", command=self.j1_plus).grid(row=0, column=2)
        self.button_j1_minus = Tk.Button(self, text="j1-", command=self.j1_minus).grid(row=0, column=3)

        self.button_j2_plus = Tk.Button(self, text="j2+", command=self.j2_plus).grid(row=1, column=2)
        self.button_j2_minus = Tk.Button(self, text="j2-", command=self.j2_minus).grid(row=1, column=3)

        self.button_j3_plus = Tk.Button(self, text="j3+", command=self.j3_plus).grid(row=2, column=2)
        self.button_j3_minus = Tk.Button(self, text="j3-", command=self.j3_minus).grid(row=2, column=3)

        self.button_j4_plus = Tk.Button(self, text="j4+", command=self.j4_plus).grid(row=3, column=2)
        self.button_j4_minus = Tk.Button(self, text="j4-", command=self.j4_minus).grid(row=3, column=3)

        self.button_j5_plus = Tk.Button(self, text="j5+", command=self.j5_plus).grid(row=4, column=2)
        self.button_j5_minus = Tk.Button(self, text="j5-", command=self.j5_minus).grid(row=4, column=3)

        self.button_j6_plus = Tk.Button(self, text="j6+", command=self.j6_plus).grid(row=5, column=2)
        self.button_j6_minus = Tk.Button(self, text="j6-", command=self.j6_minus).grid(row=5, column=3)

        # Buttons Action
        self.button_reset = Tk.Button(self, text="Open", command=self.open_gripper).grid(row=6, column=0)
        self.button_quitter = Tk.Button(self, text="Close", command=self.close_gripper).grid(row=6, column=1)
        self.button_reset = Tk.Button(self, text="Reset", command=self.reset).grid(row=6, column=2)
        self.button_quitter = Tk.Button(self, text="Quitter", command=self.quit).grid(row=6, column=3)

        self.mainloop()

    def x_plus(self):
        self.niyro_robot.arm.jog_pose([self.xyz_step, 0.0, 0.0, 0.0, 0.0, 0.0])

    def x_minus(self):
        self.niyro_robot.arm.jog_pose([-self.xyz_step, 0.0, 0.0, 0.0, 0.0, 0.0])

    def y_plus(self):
        self.niyro_robot.arm.jog_pose([0.0, self.xyz_step, 0.0, 0.0, 0.0, 0.0])

    def y_minus(self):
        self.niyro_robot.arm.jog_pose([0.0, -self.xyz_step, 0.0, 0.0, 0.0, 0.0])

    def z_plus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.00, self.xyz_step, 0.0, 0.0, 0.0])

    def z_minus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, -self.xyz_step, 0.0, 0.0, 0.0])

    def roll_plus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, 0.0, self.rpy_step, 0.0, 0.0])

    def roll_minus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, 0.0, -self.rpy_step, 0.0, 0.0])

    def pitch_plus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, 0.0, 0.0, self.rpy_step, 0.0])

    def pitch_minus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, 0.0, 0.0, -self.rpy_step, 0.0])

    def yaw_plus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.00, 0.0, 0.0, 0.0, self.rpy_step])

    def yaw_minus(self):
        self.niyro_robot.arm.jog_pose([0.0, 0.0, 0.0, 0.0, 0.0, -self.rpy_step])

    def j1_plus(self):
        self.niyro_robot.arm.jog_joints([self.joint_step, 0.0, 0.0, 0.0, 0.0, 0.0])

    def j1_minus(self):
        self.niyro_robot.arm.jog_joints([-self.joint_step, 0.0, 0.0, 0.0, 0.0, 0.0])

    def j2_plus(self):
        self.niyro_robot.arm.jog_joints([0.0, self.joint_step, 0.0, 0.0, 0.0, 0.0])

    def j2_minus(self):
        self.niyro_robot.arm.jog_joints([0.0, -self.joint_step, 0.0, 0.0, 0.0, 0.0])

    def j3_plus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.00, self.joint_step, 0.0, 0.0, 0.0])

    def j3_minus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, -self.joint_step, 0.0, 0.0, 0.0])

    def j4_plus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, 0.0, self.joint_step, 0.0, 0.0])

    def j4_minus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, 0.0, -self.joint_step, 0.0, 0.0])

    def j5_plus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, 0.0, 0.0, self.joint_step, 0.0])

    def j5_minus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, 0.0, 0.0, -self.joint_step, 0.0])

    def j6_plus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.00, 0.0, 0.0, 0.0, self.joint_step])

    def j6_minus(self):
        self.niyro_robot.arm.jog_joints([0.0, 0.0, 0.0, 0.0, 0.0, -self.joint_step])

    def reset(self):
        self.niyro_robot.arm.set_jog_control(False)
        self.niyro_robot.arm.move_joints([0.0, 0.0, 0.0, 0.0, -1.57, 0.0])
        self.niyro_robot.arm.set_jog_control(True)

    def open_gripper(self):
        self.niyro_robot.arm.set_jog_control(False)
        self.niyro_robot.tool.open_gripper()
        self.niyro_robot.arm.set_jog_control(True)

    def close_gripper(self):
        self.niyro_robot.arm.set_jog_control(False)
        self.niyro_robot.tool.close_gripper()
        self.niyro_robot.arm.set_jog_control(True)


def process(niyro_robot):
    """

    :param niyro_robot: Client
    :type niyro_robot: NiryoRobot
    """
    niyro_robot.arm.move_joints([0.0, 0.0, 0.0, 0.0, -1.57, 0.0])
    niyro_robot.arm.set_jog_control(True)

    TkinterController(niyro_robot)

    niyro_robot.arm.set_jog_control(False)
    niyro_robot.arm.set_learning_mode(True)


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto()
    # Update tool
    robot.tool.update_tool()
    # Launching main process
    process(robot)
    # Close
    robot.end()



"""
This script shows few functions you can use with PyNiryo
"""

from pyniryo2 import *

simulation_mode = True

robot_ip_address_rpi = "10.10.10.10"
robot_ip_address_simulation = "127.0.0.1"

robot_ip_address = robot_ip_address_simulation if simulation_mode else robot_ip_address_rpi

# Connecting to robot
niryo_robot = NiryoRobot(robot_ip_address)

# Trying to calibrate
niryo_robot.arm.calibrate(CalibrateMode.AUTO)
niryo_robot.arm.set_learning_mode(False)

# Getting Joints
initial_joints = niryo_robot.arm.get_joints()
print("Initial joints :\n" + str(initial_joints))

# Move Pose
niryo_robot.arm.move_pose([0.2, 0.0, 0.25, 0.0, 1.578, 0.0])

# Shift pose
niryo_robot.arm.shift_pose(RobotAxis.Y, 0.15)

# Going back to initial joints
niryo_robot.arm.move_joints(initial_joints)

niryo_robot.arm.set_learning_mode(True)

if not simulation_mode:
    # Getting hardware information
    digital_pin_array = niryo_robot.io.get_digital_io_states()
    for digital_pin in digital_pin_array:
        print(digital_pin)

    hardware_data = niryo_robot.arm.hardware_status()
    print(hardware_data)

# Close
niryo_robot.arm.set_learning_mode(True)
niryo_robot.end()



# - Imports
from pyniryo2.robot_commander import RobotCommander
from pyniryo2.utils import point_list_to_dict
from pyniryo2.objects import PoseObject

from .services import FramesServices


class Frames(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        super(Frames, self).__init__(client)

        self._services = FramesServices(self._client)

    # Main purpose
    def get_saved_dynamic_frame_list(self):
        """
        Get list of saved dynamic frames

        Example: ::

            list_frame, list_desc = robot.frames.get_saved_dynamic_frame_list()
            print(list_frame)
            print(list_desc)

        :return: list of dynamic frames name, list of description of dynamic frames
        :rtype: dictionnaire{name:description}
        """
        req = self._services.get_saved_dynamic_frame_list_request()
        response = self._services.get_frame_list_service.call(req)
        return self._services.get_saved_dynamic_frame_list_response_to_list(response)

    def get_saved_dynamic_frame(self, frame_name):
        """
        Get name, description and pose of a dynamic frame

        Example: ::

            frame = robot.frames.get_saved_dynamic_frame("default_frame")

        :param frame_name: name of the frame
        :type frame_name: str
        :return: name, description, position and orientation of a frame
        :rtype: namedtuple(name(str), description(str), position(list[float]), orientation(list[float]))
        """
        self._check_type(frame_name, str)
        req = self._services.get_dynamic_frame_from_name_request(frame_name)
        response = self._services.get_frame_from_name_service.call(req)
        self._check_result_status(response)

        return self._services.get_dynamic_frame_from_name_response_to_named_tuple(response)

    def save_dynamic_frame_from_poses(self, frame_name, description, pose_origin, pose_x, pose_y,
                                      belong_to_workspace=False):
        """
        Create a dynamic frame with 3 poses (origin, x, y)

        Example: ::

            pose_o = [0.1, 0.1, 0.1, 0, 0, 0]
            pose_x = [0.2, 0.1, 0.1, 0, 0, 0]
            pose_y = [0.1, 0.2, 0.1, 0, 0, 0]

            robot.frames.save_dynamic_frame_from_poses("name", "une description test", pose_o, pose_x, pose_y)

        :param frame_name: name of the frame
        :type frame_name: str
        :param description: description of the frame
        :type description: str
        :param pose_origin: pose of the origin of the frame
        :type pose_origin: list[float] [x, y, z, roll, pitch, yaw]
        :param pose_x: pose of the point x of the frame
        :type pose_x: list[float] [x, y, z, roll, pitch, yaw]
        :param pose_y: pose of the point y of the frame
        :type pose_y: list[float] [x, y, z, roll, pitch, yaw]
        :param belong_to_workspace: indicate if the frame belong to a workspace
        :type belong_to_workspace: boolean
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame_name, str)
        self._check_type(description, str)
        self._check_type(belong_to_workspace, bool)
        self._check_instance(pose_origin, (list, PoseObject))
        self._check_instance(pose_x, (list, PoseObject))
        self._check_instance(pose_y, (list, PoseObject))

        self._check_length(pose_origin, 6)
        self._check_length(pose_x, 6)
        self._check_length(pose_y, 6)

        pose_list = [self._args_pose_to_list(pose) for pose in (pose_origin, pose_x, pose_y)]
        req = self._services.save_dynamic_frame_from_poses_request(frame_name, description, pose_list,
                                                                   belong_to_workspace)
        response = self._services.manage_frame_service.call(req)
        self._check_result_status(response)

    def save_dynamic_frame_from_points(self, frame_name, description, point_origin, point_x, point_y,
                                       belong_to_workspace=False):
        """
        Create a dynamic frame with 3 points (origin, x, y)

        Example: ::

            point_o = [-0.1, -0.1, 0.1]
            point_x = [-0.2, -0.1, 0.1]
            point_y = [-0.1, -0.2, 0.1]

            robot.frames.save_dynamic_frame_from_points("name", "une description test", point_o, point_x, point_y)

        :param frame_name: name of the frame
        :type frame_name: str
        :param description: description of the frame
        :type description: str
        :param point_origin: origin point of the frame
        :type point_origin: list[float] [x, y, z]
        :param point_x: point x of the frame
        :type point_x: list[float] [x, y, z]
        :param point_y: point y of the frame
        :type point_y: list[float] [x, y, z]
        :param belong_to_workspace: indicate if the frame belong to a workspace
        :type belong_to_workspace: boolean
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame_name, str)
        self._check_type(description, str)
        self._check_type(belong_to_workspace, bool)
        self._check_type(point_origin, list)
        self._check_type(point_x, list)
        self._check_type(point_y, list)

        self._check_length(point_origin, 3)
        self._check_length(point_x, 3)
        self._check_length(point_y, 3)

        points_list_raw = [point_origin, point_x, point_y]
        points_list = [point_list_to_dict(point) for point in points_list_raw]

        req = self._services.save_dynamic_frame_from_points_request(frame_name, description, points_list,
                                                                    belong_to_workspace)
        response = self._services.manage_frame_service.call(req)
        self._check_result_status(response)

    def edit_dynamic_frame(self, frame_name, new_frame_name, new_description):
        """
        Modify a dynamic frame

        Example: ::

            robot.frames.edit_dynamic_frame("name", "new_name", "new description")

        :param frame_name: name of the frame
        :type frame_name: str
        :param new_frame_name: new name of the frame
        :type new_frame_name: str
        :param new_description: new description of the frame
        :type new_description: str
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame_name, str)
        self._check_type(new_frame_name, str)
        self._check_type(new_description, str)

        req = self._services.edit_dynamic_frame_request(frame_name, new_frame_name, new_description)
        response = self._services.manage_frame_service.call(req)
        self._check_result_status(response)

    def delete_saved_dynamic_frame(self, frame_name, belong_to_workspace=False):
        """
        Delete a dynamic frame

        Example: ::

            robot.frames.delete_saved_dynamic_frame("name")

        :param frame_name: name of the frame to remove
        :type frame_name: str
        :param belong_to_workspace: indicate if the frame belong to a workspace
        :type belong_to_workspace: boolean
        :return: status, message
        :rtype: (int, str)
        """
        self._check_type(frame_name, str)
        self._check_type(belong_to_workspace, bool)

        req = self._services.delete_dynamic_frame_request(frame_name, belong_to_workspace)
        response = self._services.manage_frame_service.call(req)
        self._check_result_status(response)



import roslibpy.actionlib

from pyniryo2.tool.enums import ToolCommand, ToolID
from pyniryo2.io.enums import PinID
from pyniryo2.exceptions import RobotCommandException


class ToolActions(object):

    def __init__(self, client):
        self.__client = client

        self.tool_action = None
        self.tool_action = roslibpy.actionlib.ActionClient(self.__client,
                                                           '/niryo_robot_tools_commander/action_server',
                                                           'niryo_robot_tools_commander/ToolAction')

    def __del__(self):
        if self.tool_action is not None:
            self.tool_action.cancel()
            self.tool_action.dispose()

    def get_gripper_action_goal(self, tool_id, tool_cmd, speed, max_torque_percentage, hold_torque_percentage):
        self._check_instance(tool_cmd, ToolCommand)
        self._check_instance(tool_id, ToolID)

        cmd = {'cmd_type': tool_cmd.value, 'tool_id': tool_id.value, 'speed': speed,
               'max_torque_percentage': max_torque_percentage, 'hold_torque_percentage': hold_torque_percentage}

        return roslibpy.actionlib.Goal(self.tool_action, roslibpy.Message({'cmd': cmd}))

    def get_vacuum_pump_action_goal(self, tool_cmd, tool_id):
        self._check_instance(tool_cmd, ToolCommand)

        cmd = {'cmd_type': tool_cmd.value, 'tool_id': tool_id.value,
               'activate': tool_cmd == ToolCommand.PULL_AIR_VACUUM_PUMP}

        return roslibpy.actionlib.Goal(self.tool_action, roslibpy.Message({'cmd': cmd}))

    def get_electromagnet_action_goal(self, tool_cmd, gpio=None):
        self._check_instance(tool_cmd, ToolCommand)
        if gpio is not None:
            self._check_instance(gpio, (PinID, str))

        pin = "" if gpio is None else gpio.value if isinstance(gpio, PinID) else gpio

        cmd = {'cmd_type': tool_cmd.value, 'tool_id': ToolID.ELECTROMAGNET_1.value,
               'activate': tool_cmd == ToolCommand.ACTIVATE_DIGITAL_IO, 'gpio': pin}

        return roslibpy.actionlib.Goal(self.tool_action, roslibpy.Message({'cmd': cmd}))

    def _check_instance(self, value, type_):
        if not isinstance(value, type_):
            self._raise_exception_expected_type(type_.__name__, value)

    def _raise_exception_expected_type(self, expected_type, given):
        raise RobotCommandException("Expected type: {}.\nGiven: {}".format(expected_type, given))



from pyniryo2 import *

# -- MUST Change these variables
robot_ip_address = "192.168.1.200"
workspace_name = "wks_conveyor"

# -- Can change these variables
grid_dimension = (3, 3)
conveyor_id = ConveyorID.ID_1

# -- Should Change these variables
observation_pose = PoseObject(
    x=0.20, y=0., z=0.3,
    roll=0.0, pitch=1.57, yaw=0.0,
)

center_conditioning_pose = PoseObject(
    x=0.0, y=0.25, z=0.13,
    roll=0.0, pitch=1.57, yaw=1.57,
)


def process(niryo_robot):
    """

    :type niryo_robot: NiryoRobot
    :rtype: None
    """
    max_catch_count = 6
    shape_expected = ObjectShape.CIRCLE
    color_expected = ObjectColor.RED

    catch_count = 0
    while catch_count < max_catch_count:
        # Turning conveyor on
        niryo_robot.conveyor.run_conveyor(conveyor_id)
        # Moving to observation pose
        niryo_robot.arm.move_pose(observation_pose)
        # Check if object is in the workspace
        obj_found, pos_array, shape, color = niryo_robot.vision.detect_object(workspace_name,
                                                                              shape=shape_expected,
                                                                              color=color_expected)
        if not obj_found:
            niryo_robot.wait(0.5)  # Wait to let the conveyor turn a bit
            continue
        # Stopping conveyor
        niryo_robot.conveyor.stop_conveyor(conveyor_id)
        # Making a vision pick
        obj_found, shape, colo = niryo_robot.vision.vision_pick(workspace_name,
                                                                shape=shape_expected,
                                                                color=color_expected)
        if not obj_found:  # If visual pick did not work
            continue
        # Calculating offset relative to conditioning center position
        offset_x = catch_count % grid_dimension[0] - grid_dimension[0] // 2
        offset_y = (catch_count // grid_dimension[1]) % 3 - grid_dimension[1] // 2
        place_pose = center_conditioning_pose.copy_with_offsets(0.05 * offset_x, 0.05 * offset_y)
        # Going to place
        niryo_robot.pick_place.place_from_pose(place_pose)
        catch_count += 1
    # Stopping conveyor
    niryo_robot.conveyor.stop_conveyor(conveyor_id)

    # Going to initial Observation pose
    niryo_robot.arm.move_pose(observation_pose)


if __name__ == '__main__':
    # Connect to robot
    robot = NiryoRobot(robot_ip_address)
    # Changing tool
    robot.tool.update_tool()
    # Calibrate robot if robot needs calibration
    robot.arm.calibrate_auto()
    # Launching main process
    process(robot)
    # Ending
    robot.arm.go_to_sleep()
    # Releasing connection
    robot.end()



import roslibpy
from threading import Event

from pyniryo2.exceptions import TopicException


class NiryoTopic(object):
    """
    Represent a Ros Topic instance. It supports both the request of a single value and/or callbacks.
    This class is a wrapper of roslibpy Topic instance (https://roslibpy.readthedocs.io/en/latest/reference/index.html#topics)
    """

    def __init__(self, client, topic_name, topic_type, conversion_function=None, timeout=3):
        """

        :param client: Instance of the ROS connection.
        :type client: roslibpy.Ros
        :param topic_name: Topic name.
        :type topic_name: str
        :param topic_type: Topic type.
        :type topic_type: str
        :param conversion_function: convert the response of the topic in a specific type.
        :type conversion_function: function
        :param timeout: Timeout while waiting a message.
        :type timeout: float

        """
        self.__topic_name = topic_name
        self.__topic_type = topic_type
        self.__topic = roslibpy.Topic(client, self.__topic_name, self.__topic_type)

        self.__user_callback = None

        self.__timeout = timeout

        self.__sync_topic_value = None
        self.__sync_event = Event()
        self.__sync_event.clear()

        self.__conversion_function = conversion_function

    def __del__(self):
        self.unsubscribe()

    def __call__(self):
        if self.__topic.is_subscribed:
            self.__sync_event.wait(timeout=self.__timeout)
            return self.__sync_topic_value

        self.__sync_event.clear()
        self.__topic.subscribe(self.__internal_callback)
        self.__sync_event.wait(timeout=self.__timeout)
        self.__topic.unsubscribe()
        return self.__sync_topic_value

    def __str__(self):
        return "Name: {}\nType: {}\nSubscribed: {}\nValue: {}".format(self.__topic_name, self.__topic_type,
                                                                      self.__topic.is_subscribed,
                                                                      self.__sync_topic_value)

    def __repr__(self):
        return self.__str__()

    @property
    def is_subscribed(self):
        """
        Return the topic connection status.

        :return: True if already subscribed, False otherwise.
        :rtype: Bool
        """
        return self.__topic.is_subscribed

    @property
    def value(self):
        """
        Return the last value of the topic.

        :return: The last value of the topic. The value depends on the conversion function of the topic. By default, it will be a dict.
        :rtype:
        """
        return self.__call__()

    def subscribe(self, callback):
        """
        Subscribe a callback to the topic. A TopicException will be thrown if the topic is already subscribed.

        :param callback: The callback function which is called at each incoming topic message.
        :type callback: function(dict, )
        :return: None
        :rtype: None
        """
        if not callback:
            raise TopicException("Empty callback on {} topic subscription".format(self.__topic_name))

        if not self.__topic.is_subscribed:
            self.__user_callback = callback
            self.__sync_event.clear()
            self.__topic.subscribe(self.__internal_callback)
        else:
            raise TopicException("Topic {} already subscribed".format(self.__topic_name))

    def unsubscribe(self):
        """
        Unsubscribe to the topic.

        :rtype: None
        """
        if self.__topic.is_subscribed:
            self.__topic.unsubscribe()
            self.__user_callback = None

    def publish(self, msg):
        """
        Publish a message on the topic

        :param msg: jsonified topic message content
        :type msg: dict of roslibpy.Message
        :rtype: None
        """
        self.__topic.publish(roslibpy.Message(msg) if isinstance(msg, dict) else msg)

    def __internal_callback(self, topic_value):
        """
        This function is an internal callback that stores the last value of the subject
        and calls the user's callback if it is registered.

        :param topic_value: Message returned by the topic.
        :type topic_value: dict
        :return: None
        :rtype: None
        """
        if self.__conversion_function is not None:
            self.__sync_topic_value = self.__conversion_function(topic_value)
        else:
            self.__sync_topic_value = topic_value

        self.__sync_event.set()

        if self.__user_callback:
            self.__user_callback(self.__sync_topic_value)



# Communication imports
import functools
import base64
from pyniryo2.robot_commander import RobotCommander, RobotCommandException

from .services import SoundServices
from .topics import SoundTopics
from .enums import Language


def check_ned2_version(func):
    """
    Decorator that check the robot version
    """

    @functools.wraps(func)
    def wrap(*args, **kwargs):
        robot_instance = args[0]
        if robot_instance.client.hardware_version != 'ned2':
            raise Exception("Error Code : BAD_HARDWARE_VERSION\n"
                            "Message : Wrong robot hardware version, feature only available on Ned2")

        return func(*args, **kwargs)

    return wrap


class Sound(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client):
        """
        Sound robot functions

        Example: ::

            ros_instance = NiryoRos("127.0.0.1") # Hotspot
            sound_interface = Sound(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(Sound, self).__init__(client)

        if self._client.hardware_version == 'ned2':
            self._services = SoundServices(self._client)
            self._topics = SoundTopics(self._client)
        else:
            self._services = None
            self._topics = None

    def __call__(self, *args, **kwargs):
        self.play(*args, **kwargs)

    @property
    def sounds(self):
        """
        Returns the list of available sounds in the robot

        Examples: ::

            sounds_list = sound.sounds

        :return: Returns the list of available sounds in the robot
        :rtype: list[str]
        """
        return list(self._topics.sound_database_topic().keys())

    @check_ned2_version
    def get_sounds(self):
        """
        Returns the list of available sounds in the robot

        Examples: ::

            sounds_list = sound.get_sounds()

        :return: Returns the list of available sounds in the robot
        :rtype: list[str]
        """
        return self.sounds

    @check_ned2_version
    def get_sound_duration(self, sound_name):
        """
        Get the duration of a sound in seconds

        Examples: ::

            sound_name = sound.get_sounds()[0]
            sound_duration = sound.get_sound_duration(sound_name)
            sound_duration = sound.get_sound_duration('test_sound.mp3')

        :return: Returns the duration of a sound in seconds
        :rtype: float
        """
        sounds = self._topics.sound_database_topic()
        self._check_dict_belonging(sound_name, sounds)
        return sounds[sound_name]

    @check_ned2_version
    def play(self, sound_name, wait_end=False, start_time_sec=0, end_time_sec=0):
        """
        Play a sound that as already been imported by the user on the robot. 

        Example: ::

            # If you know that the sound test_sound.wav is already imported on the robot
            sound.play_sound_user("test_sound.wav")

            # If you want to play the first sound of the ones that are already on the robot without knowing its name
            sound_name = sound.get_sounds()[0]
            sound_duration = sound.play(sound_name)

            # Waits until the sound has been fully played
            sound_duration = sound.play(sound_name, wait_end=True)

            #  Doesn't wait until the sound has been fully played
            sound_duration = sound.play(sound_name, wait_end=False)

            # Plays sound from 1.1 seconds from start to 4.3 seconds from start
            sound_duration = sound.play(sound_name, start_time_sec=1.1, end_time_sec=4.3)


        :param: sound_name: Name of the sound that will be played
        :type sound_name: str
        :param wait_end: wait for the end of the sound before exiting the function
        :type wait_end: bool
        :param start_time_sec: start the sound from this value in seconds
        :type start_time_sec: float
        :param end_time_sec: end the sound at this value in seconds
        :type end_time_sec: float
        :rtype: None
        """
        self._check_instance(sound_name, str)
        req = self._services.play_sound_request(sound_name, start_time_sec, end_time_sec, wait_end)
        resp = self._services.play_sound_service.call(req)
        self._check_result_status(resp)

    @check_ned2_version
    def stop(self):
        """
        Stop a sound being played. It will get automatically the name of the sound being played and stop it. 

        Example: ::

            self.sound.stop()

        :rtype: None
        """
        req = self._services.stop_sound_request()
        resp = self._services.stop_sound_service.call(req)
        self._check_result_status(resp)

    @property
    def state(self):
        """
        Returns the sound state client which can be used synchronously or asynchronously
        to obtain the current played sound.

        Examples: ::

            # Get last value
            sound.state()
            sound.state.value

            # Subscribe a callback
            def sound_callback(sound_name):
                print sound_name

            sound.state.subscribe(sound_callback)
            sound.state.unsubscribe()

        :return: sound state topic instance
        :rtype: NiryoTopic
        """
        return self._topics.current_sound_topic

    @property
    def volume(self):
        """
        Returns the volume state client which can be used synchronously or asynchronously
        to obtain the current volume.

        Examples: ::

            # Get last value
            sound.volume()
            sound.volume.value

            # Subscribe a callback
            def volume_callback(value):
                print value

            sound.volume.subscribe(volume_callback)
            sound.volume.unsubscribe()

        :return: volume topic instance
        :rtype: NiryoTopic
        """
        return self._topics.volume_topic

    @check_ned2_version
    def get_volume(self):
        """
        Returns the volume of the robot. The sound can be set between 0 (sound off) and 100 (sound max)

        Examples: ::

            # Get the volume of the sound
            sound.get_volume()

        :return: int8 corresponding to the volume (0: sound off, 100: sound max)
        :rtype: int8
        """
        return self._topics.volume_topic()

    @check_ned2_version
    def set_volume(self, sound_volume):
        """
        Set the volume of the robot. You can set it between 0 and 100 (0: sound off and 100: sound max).
        If you put less than 0, the volume will be set to 0.
        If you put more than 100, the volume will be set to 100.

        Example: ::

            # Set the volume to 25
            self.sound.set_volume(25)
            self.sound.play_sound_user("test_sound.wav")


        :param sound_volume: Between O and 100 (0 sound off and 100 sound maximum)
        :type sound_volume: int8
        :rtype: None
        """
        self._check_range_belonging(sound_volume, 0, 200)
        req = self._services.set_sound_volume_request(sound_volume)
        resp = self._services.set_sound_volume_service.call(req)
        self._check_result_status(resp)

    @check_ned2_version
    def delete(self, sound_name):
        """
        Delete a sound imported on the robot

        Example: ::

            self.sound.delete("test_sound.wav")


        :param sound_name: For example, test.wav
        :type sound_name: str
        :rtype: None
        """
        req = self._services.delete_sound_request(sound_name)
        resp = self._services.manage_sound_service.call(req)
        self._check_result_status(resp)

    @check_ned2_version
    def save(self, sound_name, sound_path):
        """
        Import a sound on the RaspberryPi of the robot. To do that,
        you will need the encoded data from a wav or mp3 sound.
        It is preferable to put the encoded data from the sound on a text file and directly read it from this file.
        You also need to give the name of the sound you want to import.

        Example: ::

            sound_name = "test_import_sound.wav"
            sound_path = "/home/niryo/test_sound.wav"

            ros_instance = pyniryo2.NiryoRos("10.10.10.10")
            sound = pyniryo2.Sound(ros_instance)
            sound.save(sound_name, sound_path)
            sound.play(sound_name)


        :param sound_name: For example, test.wav. Il will be the name of the sound in the robot
        :type sound_name: str
        :param sound_path: absolute path to the sound file
        :type sound_path: str
        :rtype: None
        """
        if not sound_name.endswith('.mp3') and not sound_name.endswith('.wav'):
            raise RobotCommandException("The sound name {} must end with the suffix .mp3 or .wav.".format(sound_name))

        with open(sound_path, 'rb') as f:
            sound_data = f.read()

        base64_bytes = base64.b64encode(sound_data)
        base64_message = base64_bytes.decode('ascii')

        req = self._services.import_sound_request(sound_name, base64_message)
        resp = self._services.manage_sound_service.call(req)
        self._check_result_status(resp)

    @check_ned2_version
    def say(self, text, language=Language.ENGLISH):
        """
        Use gtts (Google Text To Speech) to interpret a string as sound

        Languages available are: ::

            - English: Language. ENGLISH
            - French: Language.FRENCH
            - Spanish: Language.SPANISH
            - Mandarin: Language.MANDARIN
            - Portuguese: Language.PORTUGUESE

        Example ::

            robot.say("Hello", Language.ENGLISH)
            robot.say("Bonjour", Language.FRENCH)
            robot.say("Hola", Language.SPANISH)


        :param text: Text that needs to be spoken < 100 char
        :type text: str
        :param language: language of the text
        :type language: Language
        :rtype: None
        """
        self._check_enum_belonging(language, Language)
        self._check_type(text, str)

        req = self._services.tts_request(text, language)
        resp = self._services.tts_service.call(req)
        self._check_result_status(resp)



#!/usr/bin/env python
# coding=utf-8

from pyniryo2.io.enums import PinID, PinMode, PinState


class DigitalPinObject:
    """
    Object used to store information on digital pins
    """

    def __init__(self, name, mode, value):
        # Name
        self.pin_id = PinID(name)
        # Name
        self.name = name
        # Input or output
        self.mode = PinMode(mode)
        # True / False
        self.value = value

    def __str__(self):
        string_ret = "Name : {}".format(self.name)
        string_ret += ", Mode : {}".format(self.mode)
        string_ret += ", State : {}".format(self.value)
        return string_ret

    def __repr__(self):
        return self.__str__()


class AnalogPinObject:
    """
    Object used to store information on digital pins
    """

    def __init__(self, name, mode, value):
        # Name
        self.pin_id = PinID(name)
        # Name
        self.name = name
        # Input or output
        self.mode = PinMode(mode)
        # Tension
        self.value = value

    def __str__(self):
        string_ret = "Name : {}".format(self.name)
        string_ret += ", Mode : {}".format(self.mode)
        string_ret += ", State : {}".format(self.value)
        return string_ret

    def __repr__(self):
        return self.__str__()



# - Imports
from __future__ import print_function

# Communication imports
from pyniryo2.exceptions import RobotCommandException
from pyniryo2.robot_commander import RobotCommander
from pyniryo2.enums import RobotErrors
from pyniryo2.io.enums import PinID

from .enums import ToolID, ToolCommand
from .services import ToolServices
from .actions import ToolActions
from .topics import ToolTopics


class Tool(RobotCommander):
    def __init__(self, client):
        """
        Tool robot functions

        Example: ::

            ros_instance = NiryoRos("10.10.10.10") # Hotspot
            tool_interface = Tool(ros_instance)

        :param client: Niryo ROS client
        :type client: NiryoRos
        """
        super(Tool, self).__init__(client)

        self._services = ToolServices(self._client)
        self._topics = ToolTopics(self._client)
        self._actions = ToolActions(self._client)

        self.__action_timeout = 10

    @property
    def tool(self):
        """

        :return: The equipped tool ID
        :rtype: ToolID
        """
        return self.get_current_tool_id()

    @property
    def get_current_tool_id(self):
        """
        Returns the equipped tool Id client which can be used synchronously or asynchronously
        to obtain the equipped tool Id.
        The topic returns a attribute of the ToolID enum.

        Examples: ::

            # Get last value
            tool.get_current_tool_id()
            tool.get_current_tool_id.value

            # Subscribe a callback
            def tool_id_callback(tool_id_object):
                print tool_id_object

            tool.get_current_tool_id.subscribe(tool_id_callback)
            tool.get_current_tool_id.unsubscribe()

        :return: the equipped tool Id topic instance.
        :rtype: NiryoTopic
        """
        return self._topics.tool_id_topic

    def update_tool(self, callback=None, errback=None, timeout=None):
        """
        Update equipped tool

        Examples: ::

            # Synchronous use
            tool.update_tool()

            # Asynchronous use
            def update_tool_callback(result):
                if result["status"] < RobotErrors.SUCCESS.value:
                    print("Update failed")
                else:
                    print("Update completed with success")

            tool.update_tool(update_tool_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param errback: Callback invoked on error.
        :type errback: function
        :param timeout: Timeout for the operation, in seconds. Only used if blocking.
        :rtype: None
        """
        req = self._services.get_trigger_request()
        result = self._services.update_tool_service.call(req, callback, errback, timeout)

        if callback is None:
            self._check_result_status(result)

    def grasp_with_tool(self, callback=None):
        """
        Grasp with tool
        This action correspond to
        * Close gripper for Grippers
        * Pull Air for Vacuum pump
        * Activate for Electromagnet
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.grasp_with_tool()
            
            def tool_callback(_msg)
                print("Grasped") 
                
            tool.grasp_with_tool(callback=tool_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        tool_id = self.get_current_tool_id()

        if tool_id in (ToolID.GRIPPER_1, ToolID.GRIPPER_2, ToolID.GRIPPER_3, ToolID.GRIPPER_4):
            return self.close_gripper(callback=callback)
        elif tool_id in (ToolID.VACUUM_PUMP_1, ToolID.VACUUM_PUMP_2):
            return self.pull_air_vacuum_pump(callback=callback)
        elif tool_id == ToolID.ELECTROMAGNET_1:
            return self.activate_electromagnet(callback=callback)

    def release_with_tool(self, callback=None):
        """
        Release with tool
        This action correspond to
        * Open gripper for Grippers
        * Push Air for Vacuum pump
        * Deactivate for Electromagnet
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.release_with_tool()
            
            def tool_callback(_msg)
                print("Released")
                
            tool.release_with_tool(callback=tool_callback)

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        tool_id = self.get_current_tool_id()

        if tool_id in (ToolID.GRIPPER_1, ToolID.GRIPPER_2, ToolID.GRIPPER_3, ToolID.GRIPPER_4):
            return self.open_gripper(callback=callback)
        elif tool_id in (ToolID.VACUUM_PUMP_1, ToolID.VACUUM_PUMP_2):
            return self.push_air_vacuum_pump(callback=callback)
        elif tool_id == ToolID.ELECTROMAGNET_1:
            return self.deactivate_electromagnet(callback=callback)

    # - Gripper
    def open_gripper(self, speed=500, max_torque_percentage=100, hold_torque_percentage=30, callback=None):
        """
        Open gripper associated to the equipped gripper.
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.open_gripper()

            # Niryo One and Ned
            tool.open_gripper(speed=850)

            # Ned2
            tool.open_gripper(max_torque_percentage=100, hold_torque_percentage=50)
            
            def tool_callback(_msg)
                print("Released")
                
            tool.open_gripper(callback=tool_callback)

        :param speed: Between 100 & 1000 (only for Niryo One and Ned1)
        :type speed: int
        :param max_torque_percentage: Closing torque percentage (only for Ned2)
        :type max_torque_percentage: int
        :param hold_torque_percentage: Hold torque percentage after closing (only for Ned2)
        :type hold_torque_percentage: int
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        speed = self._transform_to_type(speed, int)
        self._check_range_belonging(speed, 1, 1000)

        max_torque_percentage = self._transform_to_type(max_torque_percentage, int)
        self._check_range_belonging(max_torque_percentage, 0, 100)

        hold_torque_percentage = self._transform_to_type(hold_torque_percentage, int)
        self._check_range_belonging(hold_torque_percentage, 0, 100)

        tool_id = self.get_current_tool_id()
        if tool_id not in [ToolID.GRIPPER_1, ToolID.GRIPPER_2, ToolID.GRIPPER_3]:
            raise RobotCommandException("Call update_tool before using the open_gripper function")

        goal = self._actions.get_gripper_action_goal(tool_id, ToolCommand.OPEN_GRIPPER,
                                                     speed, max_torque_percentage, hold_torque_percentage)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

    def close_gripper(self, speed=500, max_torque_percentage=100, hold_torque_percentage=30, callback=None):
        """
        Close gripper associated to 'gripper_id' with a speed 'speed'
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.close_gripper()

            # Niryo One and Ned
            tool.close_gripper(speed=850)

            # Ned2
            tool.close_gripper(max_torque_percentage=100, hold_torque_percentage=50)

            def tool_callback(_msg)
                print("Grasped")

            tool.close_gripper(callback=tool_callback)

        :param speed: Between 100 & 1000 (only for Niryo One and Ned1)
        :type speed: int
        :param max_torque_percentage: Opening torque percentage (only for Ned2)
        :type max_torque_percentage: int
        :param hold_torque_percentage: Hold torque percentage after opening (only for Ned2)
        :type hold_torque_percentage: int
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        speed = self._transform_to_type(speed, int)
        self._check_range_belonging(speed, 1, 1000)

        max_torque_percentage = self._transform_to_type(max_torque_percentage, int)
        self._check_range_belonging(max_torque_percentage, 0, 100)

        hold_torque_percentage = self._transform_to_type(hold_torque_percentage, int)
        self._check_range_belonging(hold_torque_percentage, 0, 100)

        tool_id = self.get_current_tool_id()
        if tool_id not in [ToolID.GRIPPER_1, ToolID.GRIPPER_2, ToolID.GRIPPER_3]:
            raise RobotCommandException("Call update_tool before using the close_gripper function")

        goal = self._actions.get_gripper_action_goal(tool_id, ToolCommand.CLOSE_GRIPPER,
                                                     speed, max_torque_percentage, hold_torque_percentage)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

        # - Vacuum

    def pull_air_vacuum_pump(self, callback=None):
        """
        Pull air of vacuum pump
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.pull_air_vacuum_pump()

            def tool_callback(_msg)
                print("Grasped")

            tool.pull_air_vacuum_pump(callback=tool_callback)


        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        tool_id = self.get_current_tool_id()
        if tool_id not in [ToolID.VACUUM_PUMP_1, ToolID.VACUUM_PUMP_2]:
            raise RobotCommandException("Call update_tool before using the pull_air_vacuum_pump function")

        goal = self._actions.get_vacuum_pump_action_goal(ToolCommand.PULL_AIR_VACUUM_PUMP, tool_id)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

    def push_air_vacuum_pump(self, callback=None):
        """
        Push air of vacuum pump
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.update_tool()
            tool.push_air_vacuum_pump()

            def tool_callback(_msg)
                print("Released")

            tool.push_air_vacuum_pump(callback=tool_callback)


        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        tool_id = self.get_current_tool_id()
        if tool_id not in [ToolID.VACUUM_PUMP_1, ToolID.VACUUM_PUMP_2]:
            raise RobotCommandException("Call update_tool before using the push_air_vacuum_pump function")

        goal = self._actions.get_vacuum_pump_action_goal(ToolCommand.PUSH_AIR_VACUUM_PUMP, tool_id)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

    # - Electromagnet

    def setup_electromagnet(self, pin_id):
        """
        Setup electromagnet on pin

        Example: ::

            tool.setup_electromagnet(PinID.GPIO_1A)
        
        :param pin_id:
        :type pin_id: PinID
        :rtype: None
        """
        self._check_instance(pin_id, (PinID, str))

        req = self._services.equip_electromagnet_service_request()
        self._services.equip_electromagnet_service.call(req)

        goal = self._actions.get_electromagnet_action_goal(ToolCommand.SETUP_DIGITAL_IO, pin_id)
        goal.send()
        self._check_result_status(goal.wait(self.__action_timeout))

    def activate_electromagnet(self, pin_id=None, callback=None):
        """
        Activate electromagnet associated to electromagnet_id on pin_id
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.setup_electromagnet(PinID.GPIO_1A)
            tool.activate_electromagnet()
            tool.activate_electromagnet(PinID.GPIO_1A)
            
            def tool_callback(_msg)
                print("Grasped") 
                
            tool.activate_electromagnet(callback=tool_callback)
            
        :param pin_id:
        :type pin_id: PinID
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        if pin_id is not None:
            self._check_instance(pin_id, (PinID, str))

        if self.get_current_tool_id() != ToolID.ELECTROMAGNET_1:
            if pin_id:
                self.setup_electromagnet(pin_id)
            else:
                raise RobotCommandException(
                    "Call setup_electromagnet before using activate_electromagnet or specify a pin ID.")

        goal = self._actions.get_electromagnet_action_goal(ToolCommand.ACTIVATE_DIGITAL_IO, pin_id)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

    def deactivate_electromagnet(self, pin_id=None, callback=None):
        """
        Deactivate electromagnet associated to electromagnet_id on pin_id
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            tool.setup_electromagnet(PinID.GPIO_1A)
            tool.deactivate_electromagnet()
            tool.deactivate_electromagnet(PinID.GPIO_1A)

            def tool_callback(_msg)
                print("Deactivated")

            tool.deactivate_electromagnet(callback=tool_callback)
            
        :param pin_id:
        :type pin_id: PinID
        :param callback: Callback invoked on successful execution.
        :type callback: function
        :rtype: None
        """
        if pin_id is not None:
            self._check_instance(pin_id, (PinID, str))

        if self.get_current_tool_id() != ToolID.ELECTROMAGNET_1:
            if pin_id:
                self.setup_electromagnet(pin_id)
            else:
                raise RobotCommandException(
                    "Call setup_electromagnet before using deactivate_electromagnet or specify a pin ID.")

        goal = self._actions.get_electromagnet_action_goal(ToolCommand.DEACTIVATE_DIGITAL_IO, pin_id)
        goal.send(result_callback=callback)

        if callback is None:
            self._check_result_status(goal.wait(self.__action_timeout))

    def enable_tcp(self, enable=True):
        """
        Enables or disables the TCP function (Tool Center Point).
        If activation is requested, the last recorded TCP value will be applied.
        The default value depends on the gripper equipped.
        If deactivation is requested, the TCP will be coincident with the tool_link.

        :param enable: True to enable, False otherwise.
        :type enable: Bool
        :rtype: None
        """
        self._check_type(enable, bool)
        req = self._services.enable_tcp_service_request(enable)
        result = self._services.enable_tcp_service.call(req)
        self._check_result_status(result)

    def set_tcp(self, *args):
        """
        Activates the TCP function (Tool Center Point)
        and defines the transformation between the tool_link frame and the TCP frame.

        Examples: ::

            tool.set_tcp(0.02, 0.0, 0.03, 0.0, 1.57, 0.0)
            tool.set_tcp([0.02, 0.0, 0.03, 0.0, 1.57, 0.0])
            tool.set_tcp(PoseObject(0.02, 0.0, 0.03, 0.0, 1.57, 0.0))

        :param args: either 6 args (1 for each coordinates) or a list of 6 coordinates or a ``PoseObject``
        :type args: Union[tuple[float], list[float], PoseObject]
        :rtype: None
        """
        pose_list = self._args_pose_to_list(*args)
        req = self._services.set_tcp_service_request(pose_list)
        result = self._services.set_tcp_service.call(req)
        self._check_result_status(result)

    def reset_tcp(self):
        """
        Reset the TCP (Tool Center Point) transformation.
        The PCO will be reset according to the tool equipped.

        :rtype: None
        """
        req = self._services.get_trigger_request()
        result = self._services.reset_tcp_service.call(req)
        self._check_result_status(result)



from enum import Enum, unique

# - Vision
@unique
class ObjectColor(Enum):
    """
    Enumeration of Colors available for image processing
    """
    RED = "RED"
    BLUE = "BLUE"
    GREEN = "GREEN"
    ANY = "ANY"


@unique
class ObjectShape(Enum):
    """
    Enumeration of Shapes available for image processing
    """
    SQUARE = "SQUARE"
    CIRCLE = "CIRCLE"
    ANY = "ANY"

@unique
class ManageWorkspace(Enum):
    """
    Enumeration of actions available for workspaces management
    """
    SAVE = 1
    SAVE_WITH_POINTS = 2
    DELETE = -1



# - Imports
import numpy as np

from pyniryo2.robot_commander import RobotCommander
from pyniryo2.objects import PoseObject

from .services import TrajectoriesServices
from .actions import TrajectoriesActions


class Trajectories(RobotCommander):
    # --- Public functions --- #
    def __init__(self, client, action_timeout=3600):
        super(Trajectories, self).__init__(client)

        self._services = TrajectoriesServices(self._client)
        self._actions = TrajectoriesActions(self._client)

        self.__action_timeout = action_timeout

    # - Main purpose
    def get_saved_trajectory(self, trajectory_name):
        """
        Get saved trajectory from robot intern storage
        Will raise error if position does not exist

        Example: ::

            trajectories.get_saved_trajectory("trajectory_01")

        :param trajectory_name: name of the trajectory
        :type trajectory_name: str
        :raises NiryoRosWrapperException: If trajectory file doesn't exist
        :return: list of [j1, j2, j3, j4, j5, j6] in rad
        :rtype: list[list[float]]
        """
        self._check_type(trajectory_name, str)

        req = self._services.get_trajectory_from_name_request(trajectory_name)
        result = self._services.get_trajectory_from_name_service.call(req)
        self._check_result_status(result)

        trajectory = self._services.trajectory_dict_to_list(result["trajectory"])
        return trajectory

    def execute_registered_trajectory(self, trajectory_name, callback=None):
        """
        Execute trajectory from Ned's memory
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            trajectories.execute_trajectory_saved("trajectory_01")

            from threading import Event
            trajectory_event = Event()
            trajectory_event.clear()

            def trajectory_callback(result):
                print(result)
                trajectory_event.set()

            trajectories.execute_trajectory_saved("trajectory_01", callback=trajectory_callback)
            trajectory_event.wait()

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :type trajectory_name: str
        :rtype: None
        """
        self._check_type(trajectory_name, str)
        req = self._services.execute_registered_trajectory_request(trajectory_name)
        response = self._services.trajectory_manager_service.call(req, callback)
        if callback is None:
            self._check_result_status(response)

    def execute_trajectory_from_poses(self, list_poses, dist_smoothing=0.0, callback=None):
        """
        Execute trajectory from list of poses
        If a callback function is not passed in parameter, the function will be blocking.
        Otherwise, the callback will be called when the execution of the function is finished.

        Examples: ::

            trajectory = [[0.3, 0.1, 0.3, 0., 0., 0., 1.],
                          [0.3, -0.1, 0.3, 0., 0., 0., 1.],
                          [0.3, -0.1, 0.4, 0., 0., 0., 1.],
                          [0.3, 0.1, 0.4, 0., 0., 0., 1.]]

            trajectories.execute_trajectory_from_poses(trajectory)
            trajectories.execute_trajectory_from_poses(trajectory, dist_smoothing=0.02)
            trajectories.execute_trajectory_from_poses([[0.3, 0.1, 0.3, 0., 0., 0., 1.], #[x,y,z,qx,qy,qz,qw]
                                                 PoseObject(0.3, -0.1, 0.3, 0., 0., 0.),
                                                 [0.3, -0.1, 0.4, 0., 0., 0.], #[x,y,z,roll,pitch,yaw]
                                                 PoseObject(0.3, 0.1, 0.4, 0., 0., 0.)])

            from threading import Event
            trajectory_event = Event()
            trajectory_event.clear()

            def trajectory_callback(result):
                print(result)
                trajectory_event.set()

            trajectories.execute_trajectory_from_poses(trajectory, callback=trajectory_callback)
            trajectory_event.wait()

        :param callback: Callback invoked on successful execution.
        :type callback: function
        :param list_poses: List of: [x,y,z,qx,qy,qz,qw] or [x,y,z,roll,pitch,yaw] or PoseObject
        :type list_poses: list[Union[tuple[float], list[float], PoseObject]]
        :param dist_smoothing: Distance from waypoints before smoothing trajectory
        :type dist_smoothing: float
        :rtype: None
        """
        self._check_range_belonging(dist_smoothing, 0.0, np.Inf)
        trajectory = self.__list_pose_to_trajectory(list_poses)
        goal = self._actions.get_execute_trajectories_goal(
            trajectory, dist_smoothing)
        goal.send(result_callback=callback)
        if callback is None:
            _result = goal.wait(self.__action_timeout)

    def save_trajectory(self, trajectory, trajectory_name, description):
        """
        Save trajectory in robot's memory

        Examples: ::

            trajectory = [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                          [1.57, 0.0, 0.0, 0.0, -1.57, 0.0],
                          [-1.57, 0.0, 0.0, 0.0, -1.57, 0.0]]

            trajectories.save_trajectory(trajectory, "trajectory_1", "test description trajectory_1")


        :param trajectory: list of joints positions the robot needs to pass by in the trajectory
        :type trajectory: list[joints]
        :param trajectory_name: name you give trajectory in the robot's memory
        :type trajectory_name: str
        :param description: description you give trajectory in the robot's memory
        :type description: str

        :rtype: None
        """

        self._check_type(trajectory_name, str)
        self._check_type(description, str)
        self._check_type(trajectory, list)
        for i, position in enumerate(trajectory):
            if len(position) != 6:
                self._raise_exception(
                    "The robot has 6 joints, position {} at index {} has only {} values".format(
                        position, i, len(position)))

        traj_message = self._services.trajectory_to_dict(trajectory)
        req = self._services.save_trajectory_request(traj_message, trajectory_name, description)
        result = self._services.trajectory_manager_service.call(req)

        self._check_result_status(result)

    def save_last_learned_trajectory(self, trajectory_name, description):
        """
        Save last executed trajectory in robot's memory

        Examples: ::

            trajectories.save_last_learned_trajectory("trajectory_1", "test description trajectory_1")

        :param trajectory_name: name you give trajectory in the robot's memory
        :type trajectory_name: str
        :param description: description you give trajectory in the robot's memory
        :type description: str
        :rtype: None
        """
        self._check_type(trajectory_name, str)
        self._check_type(description, str)

        req = self._services.save_last_learned_trajectory_request(trajectory_name, description)
        result = self._services.trajectory_manager_service.call(req)
        self._check_result_status(result)

    def delete_trajectory(self, trajectory_name):
        """
        Delete trajectory from robot's memory

        Example: ::

            if "trajectory_1" in trajectories.get_saved_trajectory_list():
                trajectories.delete_trajectory("trajectory_1")

        :type trajectory_name: str
        :rtype: None
        """
        self._check_type(trajectory_name, str)

        req = self._services.delete_trajectory_request(trajectory_name)
        response = self._services.trajectory_manager_service.call(req)
        self._check_result_status(response)

    def clean_trajectory_memory(self):
        """
        Delete all trajectories from robot's memory

        Example: ::

           trajectories.clean_trajectory_memory()

        :rtype: None
        """

        req = self._services.clean_trajectory_memory_request()
        response = self._services.trajectory_manager_service.call(req)
        self._check_result_status(response)

    def get_saved_trajectory_list(self):
        """
        Get list of trajectories' name saved in robot memory

        Example: ::

            if "trajectory_1" in trajectories.get_saved_trajectory_list():
                trajectories.delete_trajectory("trajectory_1")

        :return: list of tuple(trajectory name, trajectory definition)
        :rtype: list[tuple(str, str)]
        """
        req = self._services.get_saved_trajectory_list_request()
        response = self._services.get_trajectory_list_service.call(req)
        return list(zip(response["name_list"], response["description_list"]))

    def update_trajectory_infos(self, name, new_name, description=""):
        """
        Update the trajectory informations: name and description

        Example: ::

            trajectories.update_trajectory_infos("trajectory_1", "trajectory_2", callback="change description")

        :param name: name of the trajectory you want to change infos
        :type name: str
        :param new_name: new name you want to give to the trajectory
        :type new_name: str
        :param description: new description you want to give to the trajectory
        :type description: str
        :rtype: None
        """
        req = self._services.update_trajectory_infos_request(name, new_name, description)
        response = self._services.trajectory_manager_service.call(req)
        self._check_result_status(response)

    def __list_pose_to_trajectory(self, list_poses):
        """
        Convert a list of [x,y,z,qx,qy,qz,qw], [x,y,z,roll,pitch,yaw] and PoseObjects into a [x,y,z,qx,qy,qz,qw] list

        :param list_poses: List of: [x,y,z,qx,qy,qz,qw] or [x,y,z,roll,pitch,yaw] or PoseObject
        :type list_poses: list[Union[tuple[float], list[float], PoseObject]]
        :return: list[[x,y,z,qx,qy,qz,qw]] which represents a trajectory from the list_poses
        :rtype: list[list[float]]
        """
        self._check_type(list_poses, list)
        trajectory = []
        for pose in list_poses:
            if isinstance(pose, PoseObject):
                trajectory.append(pose.quaternion_pose)
            else:
                self._check_type(pose, list)
                if len(pose) == 7:
                    trajectory.append(self._map_list(pose[:], float))
                elif len(pose) == 6:
                    trajectory.append(PoseObject(*pose).quaternion_pose)
                else:
                    self._raise_exception("7 parameters expected in a pose [x,y,z,qx,qy,qz,qw], or  "
                                          "6 parameters expected in a pose [x,y,z,roll,pitch,yaw], or"
                                          "PoseObject expected in a pose [x,y,z,roll,pitch,yaw], "
                                          "{} given".format(pose))
        return trajectory
